# Copyright 2021-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# flake8: noqa E402
import odb
from openroad import Tech, Design

import re
import sys
import json
import locale
import inspect
import functools
from decimal import Decimal
from fnmatch import fnmatch
from typing import Callable, Dict

# -- START: Environment Fixes
try:
    locale.setlocale(locale.LC_ALL, "C.UTF-8")
except locale.Error:
    # We tried. :)
    pass
# -- END

import click
import rich
from rich.table import Table

# Re-export for subfunctions
rich
click
Table
odb

write_fn: Dict[str, Callable] = {
    "def": lambda reader, file: file and reader.design.writeDef(file),
    "odb": lambda reader, file: file and reader.design.writeDb(file),
}
auto_handled_output_opts = [f"output_{key}" for key in write_fn]


class OdbReader(object):
    def __init__(self, *args, **kwargs):
        self.ord_tech = Tech()
        self.design = Design(self.ord_tech)

        if len(args) == 1:
            db_in = args[0]
            self.design.readDb(db_in)
        elif len(args) == 2:
            lef_in, def_in = args
            if not (isinstance(lef_in, list) or isinstance(lef_in, tuple)):
                lef_in = [lef_in]
            for lef in lef_in:
                self.ord_tech.readLef(lef)
            if def_in is not None:
                self.design.readDef(def_in)

        self.config = None
        if "config_path" in kwargs and kwargs["config_path"] is not None:
            self.config = json.load(
                open(kwargs["config_path"], encoding="utf8"),
                parse_float=Decimal,
            )

        self.db = self.ord_tech.getDB()
        self.tech = self.db.getTech()
        self.chip = self.db.getChip()
        self.layers = {l.getName(): l for l in self.tech.getLayers()}
        self.libs = self.db.getLibs()
        self.cells = {}
        for lib in self.libs:
            self.cells.update({m: m for m in lib.getMasters()})
        if self.chip is not None:
            self.block = self.db.getChip().getBlock()
            self.name = self.block.getName()
            self.rows = self.block.getRows()
            self.dbunits = self.block.getDefUnits()
            self.instances = self.block.getInsts()

        busbitchars = re.escape("[]")  # TODO: Get alternatives from LEF parser
        # dividerchar = re.escape("/")  # TODO: Get alternatives from LEF parser
        self.escape_verilog_rx = re.compile(rf"([{busbitchars}])")

    def add_lef(self, new_lef):
        self.ord_tech.readLef(new_lef)

    def escape_verilog_name(self, name_in: str) -> str:
        return self.escape_verilog_rx.sub(r"\\\1", name_in)

    def _dpl(self):
        """
        The ``._dpl()`` method is EXPERIMENTAL and SHOULD NOT BE USED YET.

        Use a composite step with ``OpenROAD.DetailedPlacement``.
        """
        if self.config is None:
            raise RuntimeError("Attempted to call dpl without config file")

        cell_pad_value = int(self.config["DPL_CELL_PADDING"])
        cell_pad_side = cell_pad_value // 2
        dpl = self.design.getOpendp()
        dpl.setPaddingGlobal(cell_pad_side, cell_pad_side)
        if wildcards := self.config["CELL_PAD_EXCLUDE"]:
            for wildcard in wildcards:
                masters = [
                    self.cells[name] for name in self.cells if fnmatch(name, wildcard)
                ]
                for master in masters:
                    dpl.setPadding(master, 0, 0)
        if diode_padding := self.config["DIODE_PADDING"]:
            name, _ = self.config["DIODE_CELL"].split("/")
            master = self.cells[name]
            dpl.setPadding(master, int(diode_padding), 0)

        dpl.detailedPlacement(
            self.config["PL_MAX_DISPLACEMENT_X"],
            self.config["PL_MAX_DISPLACEMENT_Y"],
        )
        dpl.reportLegalizationStats()
        dpl.optimizeMirroring()

    def _grt(self):
        """
        The ``._grt()`` method is EXPERIMENTAL and SHOULD NOT BE USED YET.

        Use a composite step with ``OpenROAD.GlobalRouting``.
        """
        if self.config is None:
            raise RuntimeError("Attempted to call grt without config file")

        grt = self.design.getGlobalRouter()
        routing_layers = [l for l in self.layers.values() if l.getRoutingLevel() >= 1]
        for layer, adj in zip(routing_layers, self.config["GRT_LAYER_ADJUSTMENTS"]):
            grt.addLayerAdjustment(
                layer.getRoutingLevel(),
                float(adj),
            )

        min_layer_name = self.config["RT_MIN_LAYER"]
        if not min_layer_name in self.layers:
            raise RuntimeError(f"Unknown layer name '{min_layer_name}'")
        min_layer_idx = self.layers[min_layer_name].getRoutingLevel()

        max_layer_name = self.config["RT_MAX_LAYER"]
        if not max_layer_name in self.layers:
            raise RuntimeError(f"Unknown layer name '{max_layer_name}'")
        max_layer_idx = self.layers[max_layer_name].getRoutingLevel()

        min_clk_idx = min_layer_idx
        if min_clk_name := self.config["RT_CLOCK_MIN_LAYER"]:
            if not min_clk_name in self.layers:
                raise RuntimeError(f"Unknown layer name '{min_clk_name}'")
            min_clk_idx = self.layers[min_clk_name].getRoutingLevel()

        max_clk_idx = max_layer_idx
        if max_clk_name := self.config["RT_CLOCK_MAX_LAYER"]:
            if not max_clk_name in self.layers:
                raise RuntimeError(f"Unknown layer name '{max_clk_name}'")
            max_clk_idx = self.layers[max_clk_name].getRoutingLevel()

        grt.setMinRoutingLayer(min_layer_idx)
        grt.setMaxRoutingLayer(max_layer_idx)
        grt.setMinLayerForClock(min_clk_idx)
        grt.setMaxLayerForClock(max_clk_idx)
        grt.setMacroExtension(self.config["GRT_MACRO_EXTENSION"])
        grt.setOverflowIterations(self.config["GRT_OVERFLOW_ITERS"])
        grt.setAllowCongestion(self.config["GRT_ALLOW_CONGESTION"])
        grt.setVerbose(True)
        grt.globalRoute(
            True
        )  # The first variable updates guides- not sure why the default is False


def click_odb(function):
    @functools.wraps(function)
    def wrapper(input_db, input_lefs, config_path, **kwargs):
        reader = OdbReader(input_db, config_path=config_path)

        signature = inspect.signature(function)
        parameter_keys = signature.parameters.keys()

        kwargs = kwargs.copy()
        kwargs["reader"] = reader

        outputs = []
        for key, value in kwargs.items():
            if key in auto_handled_output_opts:
                id = key[7:]
                outputs.append((id, value))

        kwargs = {
            k: kwargs[k] for k in kwargs.keys() if not k in auto_handled_output_opts
        }

        if "input_db" in parameter_keys:
            kwargs["input_db"] = input_db
        if "input_lefs" in parameter_keys:
            kwargs["input_lefs"] = input_lefs

        if input_db.endswith(".def"):
            print(
                "Error: Invocation was not updated to use an odb file.", file=sys.stderr
            )
            exit(1)
        function(**kwargs)

        for format, path in outputs:
            fn = write_fn[format]
            fn(reader, path)

    for format in write_fn:
        wrapper = click.option(
            f"--output-{format}",
            default=None,
            help=f"Write {format} view",
        )(wrapper)

    wrapper = click.option(
        "-l",
        "--input-lef",
        "input_lefs",
        default=(),
        help="LEF file needed to have a proper view of the DEF files",
        multiple=True,
    )(wrapper)
    wrapper = click.option(
        "--step-config",
        "config_path",
        type=click.Path(exists=True, dir_okay=False, file_okay=True, readable=True),
        required=False,
    )(wrapper)
    wrapper = click.argument("input_db")(wrapper)

    return wrapper



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class DesignFormatObject:
    """
    Metadata about the various possible text or binary representations (views)
    of any design.

    For example, ``DesignFormat.NETLIST.value`` has the metadata for Netlist
    views.

    :param id: A lowercase alphanumeric identifier for the design format.
        Some IDs in OpenLane 2.X use dashes. This is an inconsistency that will
        be addressed in the next major version of OpenLane as it would be a
        breaking change.
    :param extension: The file extension for designs saved in this format.
    :param name: A human-readable name for this design format.
    :param folder_override: The subdirectory when
        :meth:`openlane.state.State.save_snapshot` is called on a state. If
        unset, the value for ``id`` will be used.
    :param multiple: Whether this view may have multiple files (typically, files
        that are different across multiple corners or similar.)
    """

    id: str
    extension: str
    name: str
    folder_override: Optional[str] = None
    multiple: bool = False

    @property
    def folder(self) -> str:
        return self.folder_override or self.id


class DesignFormat(Enum):
    """
    An `enumeration <https://docs.python.org/3/library/enum.html>`_ of a number
    of :class:`openlane.state.DesignFormatObject`\\s representing the various
    possible text or binary representations (views) supported by OpenLane
    states.

    Members of this enumeration are used as the keys of
    :class:`openlane.state.State` objects.
    """

    NETLIST: DesignFormatObject = DesignFormatObject(
        "nl",
        "nl.v",
        "Verilog Netlist",
    )
    POWERED_NETLIST: DesignFormatObject = DesignFormatObject(
        "pnl",
        "pnl.v",
        "Powered Verilog Netlist",
    )
    POWERED_NETLIST_SDF_FRIENDLY: DesignFormatObject = DesignFormatObject(
        "pnl-sdf-friendly",
        "pnl-sdf.v",
        "Powered Verilog Netlist For SDF Simulation (Without Fill Cells)",
        folder_override="pnl",
    )
    POWERED_NETLIST_NO_PHYSICAL_CELLS: DesignFormatObject = DesignFormatObject(
        "pnl-npc",
        "pnl-npc.v",
        "Powered Verilog Netlist Without Physical Cells (Fill Cells and Diode Cells)",
        folder_override="pnl",
    )

    DEF: DesignFormatObject = DesignFormatObject(
        "def",
        "def",
        "Design Exchange Format",
    )
    LEF: DesignFormatObject = DesignFormatObject(
        "lef",
        "lef",
        "Library Exchange Format",
    )
    OPENROAD_LEF: DesignFormatObject = DesignFormatObject(
        "openroad-lef",
        "openroad.lef",
        "Library Exchange Format Generated by OpenROAD",
        folder_override="lef",
    )
    ODB: DesignFormatObject = DesignFormatObject(
        "odb",
        "odb",
        "OpenDB Database",
    )

    SDC: DesignFormatObject = DesignFormatObject(
        "sdc",
        "sdc",
        "Design Constraints",
    )
    SDF: DesignFormatObject = DesignFormatObject(
        "sdf",
        "sdf",
        "Standard Delay Format",
        multiple=True,
    )
    SPEF: DesignFormatObject = DesignFormatObject(
        "spef",
        "spef",
        "Standard Parasitics Extraction Format",
        multiple=True,  # nom, min, max, ...
    )
    LIB: DesignFormatObject = DesignFormatObject(
        "lib",
        "lib",
        "LIB Timing Library Format",
        multiple=True,
    )
    SPICE: DesignFormatObject = DesignFormatObject(
        "spice",
        "spice",
        "Simulation Program with Integrated Circuit Emphasis",
    )

    MAG: DesignFormatObject = DesignFormatObject(
        "mag",
        "mag",
        "Magic VLSI View",
    )

    GDS: DesignFormatObject = DesignFormatObject(
        "gds",
        "gds",
        "GDSII Stream",
    )
    MAG_GDS: DesignFormatObject = DesignFormatObject(
        "mag_gds",
        "magic.gds",
        "GDSII Stream (Magic)",
    )
    KLAYOUT_GDS: DesignFormatObject = DesignFormatObject(
        "klayout_gds",
        "klayout.gds",
        "GDSII Stream (KLayout)",
    )

    JSON_HEADER: DesignFormatObject = DesignFormatObject(
        "json_h",
        "h.json",
        "Design JSON Header File",
    )
    VERILOG_HEADER: DesignFormatObject = DesignFormatObject(
        "vh",
        "vh",
        "Verilog Header",
    )

    def __str__(self) -> str:
        return self.value.id

    @staticmethod
    def by_id(id: str) -> Optional["DesignFormat"]:
        return _designformat_by_id.get(id)


_designformat_by_id: Dict[str, "DesignFormat"] = {
    format.value.id: format for format in DesignFormat
}



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shlex
import shutil
import datetime
import functools
import subprocess
from functools import partial
from typing import IO, Any, Dict, Optional, Sequence, Union

from click import pass_context, Context, argument
from cloup import (
    group,
    option,
    command,
    Path,
)

from .step import Step, StepError, StepException
from ..logging import info, err, warn
from ..flows import cloup_flow_opts
from ..__version__ import __version__
from ..common.cli import formatter_settings
from ..common import mkdirp, Toolbox, get_openlane_root


def load_step_from_inputs(
    ctx: Context,
    id: Optional[str],
    config: str,
    state_in: str,
    pdk_root: Optional[str] = None,
) -> Step:
    Target = Step
    if id is not None:
        if Found := Step.factory.get(id):
            Target = Found
        else:
            err(
                f"No step registered with id '{id}'. Ensure all relevant plugins are installed."
            )
            ctx.exit(-1)

    return Target.load(
        config=config,
        state_in=state_in,
        pdk_root=pdk_root,
    )


o = partial(option, show_default=True)


@command(formatter_settings=formatter_settings)
@o(
    "-o",
    "--output",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    help="The directory to store artifacts from step execution in",
    default=os.path.join(
        os.getcwd(),
        datetime.datetime.now().astimezone().strftime("STEP_RUN_%Y-%m-%d_%H-%M-%S"),
    ),
)
@o(
    "--id",
    type=str,
    required=False,
    help="The ID for the step. Can be omitted if the configuration object has the ID in the key-path .meta.step.",
)
@o(
    "-c",
    "--config",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=True,
    help="A step-specific config.json file",
)
@o(
    "-i",
    "--state-in",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=False,
)
@o(
    "--pdk-root",
    type=Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
    ),
    is_eager=True,
    default=os.environ.pop("PDK_ROOT", None),
    help="Use this folder as the PDK root, if running a reproducible that doesn't include the PDK.",
)
@cloup_flow_opts(
    config_options=False,
    run_options=False,
    sequential_flow_controls=False,
    jobs=False,
    accept_config_files=False,
)
@pass_context
def run(ctx, output, state_in, config, id, pdk_root, pdk, scl):
    """
    Runs a step using a step-specific configuration object and an input state.

    Useful for re-running a step that has already run, or running
    filesystem-independent reproducibles.
    """

    step = load_step_from_inputs(ctx, id, config, state_in, pdk_root)

    if step.config.meta.openlane_version != __version__:
        warn(
            "OpenLane version being used is different from the version this step was originally run with. Procceed with caution."
        )

    mkdirp(output)
    toolbox_dir = os.path.join(output, "toolbox_tmp")
    try:
        step.start(
            toolbox=Toolbox(toolbox_dir),
            step_dir=output,
        )
    except StepException as e:
        err("An unexpected error occurred while executing your step:")
        err(e)
        ctx.exit(-1)
    except StepError as e:
        err("An error occurred while executing your step:")
        err(e)
        ctx.exit(-1)


@command(formatter_settings=formatter_settings)
@o(
    "-o",
    "--output",
    type=Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
    ),
    help="Ejected run script",
    default="run.sh",
)
@o(
    "--id",
    type=str,
    required=False,
    help="The ID for the step. Can be omitted if the configuration object has the ID in the key-path .meta.step.",
)
@o(
    "-c",
    "--config",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=True,
    help="A step-specific config.json file",
)
@o(
    "-i",
    "--state-in",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=False,
)
@pass_context
def eject(ctx, output, state_in, config, id):
    """
    For steps that rely on underlying utilities using a subprocess, this scripts
    "ejects" OpenLane and just returns a shell script that runs this subprocess.

    This is useful for:

    * OpenLane developers and maintainers reporting issues to original tool
      developers
    * Advanced users who are sure that the issue is not OpenLane-specific and
      would like to skip reporting an issue with OpenLane first
    """

    step = load_step_from_inputs(ctx, id, config, state_in)

    if step.config.meta.openlane_version != __version__:
        warn(
            "OpenLane version being used is different from the version this step was originally run with. Procceed with caution."
        )

    toolbox_dir = os.path.join(".", "toolbox_tmp")

    found_cmd: Optional[Sequence[Union[str, os.PathLike]]] = None
    found_env: Optional[Dict[str, Any]] = None
    found_stdin_data: Optional[Union[str, bytes]] = None

    class Stop(Exception):
        pass

    def popen_substitute(
        cmd: Sequence[Union[str, os.PathLike]],
        env: Optional[Dict[str, Any]] = None,
        stdin: Optional[IO[Any]] = None,
        *args,
        **kwargs,
    ) -> subprocess.Popen:
        nonlocal found_env, found_cmd, found_stdin_data
        found_cmd = cmd
        found_env = env
        if found_stdin := stdin:
            found_stdin_data = found_stdin.read()
        raise Stop()

    step.run_subprocess = functools.partial(
        step.run_subprocess,
        _popen_callable=popen_substitute,
    )

    try:
        step.start(
            toolbox=Toolbox(toolbox_dir),
            step_dir=".",
        )
    except Stop:
        pass
    except Exception as e:
        info("An error occurred while attempting to execute the step:")
        err(e)
        info("This may affect the ejection process.")

    if found_cmd is None:
        err(
            "Could not eject: The step did not successfully invoke a subprocess using run_subprocess."
        )
        exit(-1)

    canon_scripts_dir = os.path.join(get_openlane_root(), "scripts")
    target_scripts_dir = os.path.join(".", "scripts")

    try:
        shutil.rmtree(target_scripts_dir)
    except FileNotFoundError:
        pass

    shutil.copytree(canon_scripts_dir, target_scripts_dir)
    if chmod := shutil.which("chmod"):
        # Nix's Files aren't writeable
        subprocess.check_call([chmod, "-R", "755", target_scripts_dir])

    current_env = os.environ
    filtered_env = {
        "STEP_DIR": ".",
        "SCRIPTS_DIR": target_scripts_dir,
    }
    if found_env is not None:
        for key, value in found_env.items():
            if (
                value == current_env.get(key)
                or key in filtered_env
                or key in ["PATH", "PYTHONPATH"]
            ):
                continue
            if os.path.isabs(value) and os.path.exists(value):
                if value.startswith(canon_scripts_dir):
                    value = value.replace(canon_scripts_dir, target_scripts_dir)
            filtered_env[key] = value

    cat_in = ""
    if found_stdin_data:
        mode = "wb"
        if isinstance(found_stdin_data, str):
            mode = "w"
        with open("STDIN", mode) as f:
            f.write(found_stdin_data)
        cat_in = "cat STDIN | "

    found_cmd_filtered = []
    for cmd in found_cmd:
        cmd = str(cmd).replace(canon_scripts_dir, target_scripts_dir)
        found_cmd_filtered.append(cmd)

    with open(output, "w", encoding="utf8") as f:
        f.write("#!/bin/sh\n")
        for key, value in filtered_env.items():
            f.write(f"export {key}={shlex.quote(str(value))}\n")
        f.write("\n")
        f.write(cat_in)
        f.write(shlex.join([str(e) for e in found_cmd_filtered]))
        f.write("\n")

    if hasattr(os, "chmod"):
        os.chmod(output, 0o755)

    info("Ejected successfully.")


@command(formatter_settings=formatter_settings)
@o(
    "-o",
    "--output",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    help="The output directory for the reproducible.",
    default="reproducible",
)
@o(
    "-d",
    "--step-dir",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    help="The step directory from which to create the reproducible. If provided, --config and the input state can be omitted, and vice versa.",
    default=None,
)
@o(
    "--id",
    type=str,
    required=False,
    help="The ID for the step. Can be omitted if the configuration object has the ID in the key-path .meta.step.",
)
@o(
    "-c",
    "--config",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=False,
    help="A step-specific config.json file",
)
@o(
    "-i",
    "--state-in",
    type=Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
    ),
    required=False,
)
@o(
    "--include-pdk/--no-include-pdk",
    type=bool,
    default=True,
)
@o(
    "--flatten/--no-flatten",
    type=bool,
    default=False,
)
@pass_context
@argument(
    "step_dir_arg",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    default=None,
    required=False,
    nargs=1,
)
def create_reproducible(
    ctx,
    output,
    step_dir,
    step_dir_arg,
    id,
    config,
    state_in,
    include_pdk,
    flatten,
):
    """
    Creates a filesystem-independent step reproducible.

    The input can be either:

    * Both a configuration object (--config) and an input state (--state-in)

    * A step directory (--step-dir) generated from a previous run
      * If not provided, the fallbacks are, in order of priority:
        * The first non-flag argument
        * The current working directory

    These reproducibles are filesystem-independent, i.e. they can be run
    on any computer that has the appropriate version of OpenLane 2 installed
    (as well as the underlying utility for that specific step.)

    The reproducible will report an error if OpenLane is not installed and will
    emit a warning if the installed version of OpenLane mismatches the one
    declared in the config file.
    """
    step_dir = step_dir or step_dir_arg or os.getcwd()

    if step_dir is None:
        if config is None or state_in is None:
            err("Either --step-dir or both --config and --state-in must be provided.")
            ctx.exit(-1)
        elif None in [config, state_in]:
            err(
                "Both --config and --state-in must be provided if the --step-dir is not provided."
            )
            ctx.exit(-1)
    else:
        if config is None:
            config = os.path.join(step_dir, "config.json")
        if state_in is None:
            state_in = os.path.join(step_dir, "state_in.json")

    step = load_step_from_inputs(ctx, id, config, state_in)
    step.create_reproducible(output, include_pdk, flatten=flatten)


@command(formatter_settings=formatter_settings, hidden=True)
@o(
    "-d",
    "--step-dir",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    help="The step directory from which to create the test. If provided, --config and the input state can be omitted, and vice versa.",
    default=None,
)
@o(
    "--output",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    default=None,
)
@argument(
    "step_dir_arg",
    type=Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
    default=None,
    required=False,
    nargs=1,
)
@pass_context
def create_test(ctx, step_dir, step_dir_arg, output):
    step_dir = step_dir or step_dir_arg or os.getcwd()
    config = os.path.join(step_dir, "config.json")
    state_in = os.path.join(step_dir, "state_in.json")
    if output is None:
        output = os.path.join(step_dir, "test")

    step = load_step_from_inputs(ctx, None, config, state_in)
    step.create_reproducible(output, include_pdk=False, flatten=True)
    os.remove(os.path.join(output, "run_ol.sh"))
    if os.path.exists(os.path.join(output, "base.sdc")):
        os.remove(os.path.join(output, "base.sdc"))


@group(formatter_settings=formatter_settings)
def cli():
    """
    Try 'python3 -m openlane.steps COMMAND --help' for help with a specific
    command.
    """
    pass


cli.add_command(run)
cli.add_command(eject)
cli.add_command(create_reproducible)
cli.add_command(create_test)

if __name__ == "__main__":
    cli()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .metric import Metric, sum_aggregator, min_aggregator, max_aggregator


# Area and Counts
Metric(
    "design__core__area",
    higher_is_better=False,
)
Metric(
    "design__die__area",
    higher_is_better=False,
)
Metric(
    "design__instance__area",
    higher_is_better=False,
)
Metric(
    "design__instance__count",
    higher_is_better=False,
)

# Power
Metric(
    "ir__drop__avg",
    higher_is_better=False,
)
Metric(
    "ir__drop__worst",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "ir__voltage__worst",
    aggregator=min_aggregator,
    higher_is_better=True,
)
Metric(
    "design_powergrid__drop__average",
    higher_is_better=False,
)
Metric(
    "design_powergrid__drop__worst",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "design_powergrid__voltage__worst",
    aggregator=min_aggregator,
    higher_is_better=True,
)
Metric(
    "power__internal__total",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "power__leakage__total",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "power__switching__total",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "power__total",
    aggregator=sum_aggregator,
    higher_is_better=False,
)

# Timing
Metric(
    "timing__hold_vio__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "timing__hold_r2r_vio__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "timing__setup_vio__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "timing__setup_r2r_vio__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "timing__hold__ws",
    aggregator=min_aggregator,
    higher_is_better=True,
)
Metric(
    "timing__hold_r2r__ws",
    aggregator=min_aggregator,
    higher_is_better=True,
)
Metric(
    "timing__setup__ws",
    aggregator=min_aggregator,
    higher_is_better=True,
)

Metric(
    "timing__setup_r2r__ws",
    aggregator=min_aggregator,
    higher_is_better=True,
)
Metric(
    "timing__hold__wns",
    aggregator=min_aggregator,
    higher_is_better=True,
    critical=True,
)
Metric(
    "timing__setup__wns",
    aggregator=min_aggregator,
    higher_is_better=True,
    critical=True,
)
Metric(
    "timing__hold__tns",
    aggregator=min_aggregator,
    higher_is_better=True,
    critical=True,
)
Metric(
    "timing__unannotated_net__count",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "timing__unannotated_net_filtered__count",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "timing__setup__tns",
    aggregator=min_aggregator,
    higher_is_better=True,
    critical=True,
)
Metric(
    "clock__skew__worst_hold",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "clock__skew__worst_setup",
    aggregator=min_aggregator,
    higher_is_better=True,
)

# Constraint Violation
Metric(
    "design__max_slew_violation__count",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "design__max_fanout_violation__count",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "design__max_cap_violation__count",
    aggregator=max_aggregator,
    higher_is_better=False,
)

# Placement and Routing
Metric(
    "route__wirelength",
    aggregator=sum_aggregator,
    higher_is_better=False,
    dont_aggregate=["iter"],
)
Metric(
    "route__wirelength__estimated",
    aggregator=sum_aggregator,
    higher_is_better=False,
    dont_aggregate=["iter"],
)
Metric(
    "route__wirelength__max",
    aggregator=max_aggregator,
    higher_is_better=False,
    dont_aggregate=["iter"],
)
Metric(
    "route__antenna_violation__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    dont_aggregate=["iter"],
)
Metric(
    "design__instance__displacement__max",
    aggregator=max_aggregator,
    higher_is_better=False,
)
Metric(
    "design__instance__displacement__total",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "design__instance__utilization",
    higher_is_better=True,
)

# Potential Issues
Metric(
    "design__lint_warning__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "design__lint_error__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lint_timing_construct__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "antenna__violating__nets",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "antenna__violating__pins",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "design__instance_unmapped__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__disconnected_pin__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__inferred_latch__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "design__violations",
    aggregator=sum_aggregator,
    higher_is_better=False,
)
Metric(
    "design__xor_difference__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "route__drc_errors",
    aggregator=sum_aggregator,
    higher_is_better=False,
    dont_aggregate=["iter"],
    critical=True,
)
Metric(
    "magic__drc_error__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "magic__illegal_overlap__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "synthesis__check_error__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_error__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_device_difference__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_net_difference__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_property_fail__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_unmatched_device__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_unmatched_net__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)
Metric(
    "design__lvs_unmatched_pin__count",
    aggregator=sum_aggregator,
    higher_is_better=False,
    critical=True,
)



# Copyright 2020-2024 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#  Parts of this file adapted from https://github.com/YosysHQ/yosys/blob/master/techlibs/common/synth.cc
#
#  Copyright (C) 2012  Claire Xenia Wolf <claire@yosyshq.com>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import sys
import json
import shutil

import click

from ys_common import ys
from construct_abc_script import ABCScriptCreator


def openlane_proc(d: ys.Design, report_dir: str):
    d.run_pass("proc_clean")  # Clean up unused procedures
    d.run_pass("proc_rmdead")  # Remove dead procedures
    d.run_pass("proc_prune")  # Prune unused parts of procedures
    d.run_pass("proc_init")  # Initialize procedures
    d.run_pass("proc_arst")  # Analyze reset signals in procedures
    d.run_pass("proc_rom")  # Map ROMs within procedures
    d.run_pass("proc_mux")  # Optimize multiplexers within procedures
    d.tee(
        "proc_dlatch", o=os.path.join(report_dir, "latch.rpt")
    )  # Report latches after procedure processing
    d.run_pass("proc_dff")  # Analyze flip-flops within procedures
    d.run_pass("proc_memwr")  # Analyze memory writes within procedures
    d.run_pass("proc_clean")  # Clean up after procedure processing
    d.tee("check", o=os.path.join(report_dir, "pre_synth_chk.rpt"))
    d.run_pass("opt_expr")  # Optimize expressions


def openlane_synth(d, top, flatten, report_dir, *, booth=False, abc_dff=False):
    d.run_pass("hierarchy", "-check", "-top", top, "-nokeep_prints", "-nokeep_asserts")
    openlane_proc(d, report_dir)

    if flatten:
        d.run_pass("flatten")  # Flatten the design hierarchy

    d.run_pass("opt_expr")  # Optimize expressions again

    d.run_pass("opt_clean")  # Clean up after optimization

    # Perform various logic optimization passes
    d.run_pass("opt", "-nodffe", "-nosdff")  # Optimize logic excluding flip-flops
    d.run_pass("fsm")  # Identify and optimize finite state machines
    d.run_pass("opt")  # Additional logic optimization
    d.run_pass("wreduce")  # Reduce logic using algebraic rewriting
    d.run_pass("peepopt")  # Perform local peephole optimization
    d.run_pass("opt_clean")  # Clean up after optimization
    if booth:
        d.run_pass("booth")
    d.run_pass("alumacc")  # Optimize arithmetic logic unitsb
    d.run_pass("share")  # Share logic across the design
    d.run_pass("opt")  # More logic optimization

    # Memory optimization
    d.run_pass("memory", "-nomap")  # Analyze memories but don't map them yet
    d.run_pass("opt_clean")  # Clean up after memory analysis

    # Perform more aggressive optimization with faster runtime
    d.run_pass("opt", "-fast", "-full")  # Fast and comprehensive optimization

    # Technology mapping
    d.run_pass("memory_map")  # Map memories to standard cells
    d.run_pass("opt", "-full")  # More optimization after memory mapping
    d.run_pass("techmap")  # Map logic to standard cells from the technology library
    d.run_pass("opt", "-fast")  # Fast optimization after technology mapping
    d.run_pass("opt", "-fast")  # More fast optimization

    d.run_pass(
        "abc", "-fast", *(["-dff"] if abc_dff else [])
    )  # Run ABC with fast settings
    d.run_pass("opt", "-fast")  # MORE fast optimization

    # Checks and Stats
    d.run_pass("hierarchy", "-check", "-top", top, "-nokeep_prints", "-nokeep_asserts")
    d.run_pass("check")
    d.run_pass("stat")


@click.command()
@click.option("--output", type=click.Path(exists=False, dir_okay=False), required=True)
@click.option("--config-in", type=click.Path(exists=True), required=True)
@click.option("--extra-in", type=click.Path(exists=True), required=True)
@click.option("--lighter-dff-map", type=click.Path(exists=True), required=False)
def synthesize(
    output,
    config_in,
    extra_in,
    lighter_dff_map,
):
    config = json.load(open(config_in))
    extra = json.load(open(extra_in))

    includes = config.get("VERILOG_INCLUDE_DIRS") or []
    defines = (config.get("VERILOG_DEFINES") or []) + [
        f"PDK_{config['PDK']}",
        f"SCL_{config['STD_CELL_LIBRARY']}",
        "__openlane__",
        "__pnr__",
    ]

    blackbox_models = extra["blackbox_models"]
    libs = extra["libs_synth"]

    d = ys.Design()

    step_dir = os.path.dirname(output)
    report_dir = os.path.join(step_dir, "reports")
    os.makedirs(report_dir, exist_ok=True)

    d.add_blackbox_models(blackbox_models, includes=includes, defines=defines)

    clock_period = config["CLOCK_PERIOD"] * 1000  # ns -> ps

    # ABC only supports these two:
    # https://github.com/YosysHQ/abc/blob/28d955ca97a1c4be3aed4062aec0241a734fac5d/src/map/scl/sclUtil.c#L257
    sdc_path = os.path.join(step_dir, "synthesis.abc.sdc")
    with open(sdc_path, "w") as f:
        print(f"set_driving_cell {config['SYNTH_DRIVING_CELL']}", file=f)
        print(f"set_load {config['OUTPUT_CAP_LOAD']}", file=f)

    ys.log(f"[INFO] Using SDC file '{sdc_path}' for ABC…")

    if verilog_files := config.get("VERILOG_FILES"):
        d.read_verilog_files(
            verilog_files,
            top=config["DESIGN_NAME"],
            synth_parameters=config["SYNTH_PARAMETERS"] or [],
            includes=includes,
            defines=defines,
            use_synlig=config["USE_SYNLIG"],
            synlig_defer=config["SYNLIG_DEFER"],
        )
    elif vhdl_files := config.get("VHDL_FILES"):
        d.run_pass("plugin", "-i", "ghdl")
        d.run_pass("ghdl", *vhdl_files, "-e", config["DESIGN_NAME"])
    else:
        print(
            "Script called inappropriately: config must include either VERILOG_FILES or VHDL_FILES.",
            file=sys.stderr,
        )
        exit(1)

    d.run_pass(
        "hierarchy",
        "-check",
        "-top",
        config["DESIGN_NAME"],
        "-nokeep_prints",
        "-nokeep_asserts",
    )
    d.run_pass("rename", "-top", config["DESIGN_NAME"])
    d.run_pass("select", "-module", config["DESIGN_NAME"])
    try:
        d.run_pass(
            "show", "-format", "dot", "-prefix", os.path.join(step_dir, "hierarchy")
        )
    except Exception:
        pass
    d.run_pass("select", "-clear")

    lib_arguments = []
    for lib in libs:
        lib_arguments.extend(["-liberty", lib])

    if config["SYNTH_ELABORATE_ONLY"]:
        openlane_proc(d, report_dir)
        if config["SYNTH_ELABORATE_FLATTEN"]:
            d.run_pass("flatten")
        d.run_pass("setattr", "-set", "keep", "1")
        d.run_pass("splitnets")
        d.run_pass("opt_clean", "-purge")
        d.tee("check", o=os.path.join(report_dir, "chk.rpt"))
        d.tee("stat", "-json", *lib_arguments, o=os.path.join(report_dir, "stat.json"))
        d.tee("stat", *lib_arguments, o=os.path.join(report_dir, "stat.rpt"))

        d.run_pass(
            "write_verilog",
            "-noattr",
            "-noexpr",
            "-nohex",
            "-nodec",
            "-defparam",
            output,
        )
        d.run_pass("write_json", f"{output}.json")
        exit(0)

    if config["SYNTH_TRISTATE_MAP"] is not None:
        d.run_pass("tribuf")

    adder_type = config["SYNTH_ADDER_TYPE"]
    if adder_type not in ["YOSYS", "FA"]:
        if mapping := config[f"SYNTH_{adder_type}_MAP"]:
            ys.log(f"[INFO] Applying {adder_type} mapping from '{mapping}'…")
            d.run_pass("techmap", "-map", mapping)

    if mapping := lighter_dff_map:
        ys.log(f"[INFO] Using Lighter with mapping '{mapping}'…")
        d.run_pass("plugin", "-i", "lighter")
        d.run_pass("reg_clock_gating", "-map", mapping)

    openlane_synth(
        d,
        config["DESIGN_NAME"],
        not config["SYNTH_NO_FLAT"],
        report_dir,
        booth=config["SYNTH_MUL_BOOTH"],
        abc_dff=config["SYNTH_ABC_DFF"],
    )

    d.run_pass("delete", "t:$print")
    d.run_pass("delete", "t:$assert")

    try:
        d.run_pass(
            "show",
            "-format",
            "dot",
            "-prefix",
            os.path.join(step_dir, "primitive_techmap"),
        )
    except Exception:
        pass

    d.run_pass("opt")
    d.run_pass("opt_clean", "-purge")

    d.tee(
        "stat", "-json", *lib_arguments, o=os.path.join(report_dir, "pre_techmap.json")
    )
    d.tee("stat", *lib_arguments, o=os.path.join(report_dir, "pre_techmap.rpt"))

    if tristate_mapping := config["SYNTH_TRISTATE_MAP"]:
        ys.log(f"[INFO] Applying tri-state buffer mapping from '{tristate_mapping}'…")
        d.run_pass("techmap", "-map", tristate_mapping)
        d.run_pass("simplemap")
    if fa_mapping := config["SYNTH_FA_MAP"]:
        if adder_type == "FA":
            ys.log(f"[INFO] Applying full-adder mapping from '{fa_mapping}'…")
            d.run_pass("techmap", "-map", fa_mapping)
    if latch_mapping := config["SYNTH_LATCH_MAP"]:
        ys.log(f"[INFO] Applying latch mapping from '{latch_mapping}'…")
        d.run_pass("techmap", "-map", latch_mapping)
        d.run_pass("simplemap")
    if extra_mapping := config["SYNTH_EXTRA_MAPPING_FILE"]:
        ys.log(f"[INFO] Applying extra mappings from '{extra_mapping}'…")
        d.run_pass("techmap", "-map", extra_mapping)

    dfflibmap_args = []
    for lib in libs:
        dfflibmap_args.extend(["-liberty", lib])
    d.run_pass("dfflibmap", *dfflibmap_args)

    d.tee("stat", "-json", *lib_arguments, o=os.path.join(report_dir, "post_dff.json"))
    d.tee("stat", *lib_arguments, o=os.path.join(report_dir, "post_dff.rpt"))

    script_creator = ABCScriptCreator(config)

    def run_strategy(d):
        abc_script = script_creator.generate_abc_script(
            step_dir,
            config["SYNTH_STRATEGY"],
        )
        ys.log(f"[INFO] Using generated ABC script '{abc_script}'…")
        d.run_pass(
            "abc",
            "-script",
            abc_script,
            "-D",
            f"{clock_period}",
            "-constr",
            sdc_path,
            "-showtmp",
            *lib_arguments,
            *(["-dff"] if config["SYNTH_ABC_DFF"] else []),
        )

        d.run_pass("setundef", "-zero")
        d.run_pass(
            "hilomap",
            "-hicell",
            *config["SYNTH_TIEHI_CELL"].split("/"),
            "-locell",
            *config["SYNTH_TIELO_CELL"].split("/"),
        )

        if config["SYNTH_SPLITNETS"]:
            d.run_pass("splitnets")
            d.run_pass("opt_clean", "-purge")

        if config["SYNTH_DIRECT_WIRE_BUFFERING"]:
            d.run_pass("insbuf", "-buf", *config["SYNTH_BUFFER_CELL"].split("/"))

        d.tee("check", o=os.path.join(report_dir, "chk.rpt"))
        d.tee("stat", "-json", *lib_arguments, o=os.path.join(report_dir, "stat.json"))
        d.tee("stat", *lib_arguments, o=os.path.join(report_dir, "stat.rpt"))

        if config["SYNTH_AUTONAME"]:
            # Generate public names for the various nets, resulting in very long
            # names that include the full hierarchy, which is preferable to the
            # internal names that are simply sequential numbers such as `_000019_`.
            # Renamed net names can be very long, such as:
            #     manual_reset_gf180mcu_fd_sc_mcu7t5v0__dffq_1_Q_D_gf180mcu_ \
            #     fd_sc_mcu7t5v0__nor3_1_ZN_A1_gf180mcu_fd_sc_mcu7t5v0__aoi21_ \
            #     1_A2_A1_gf180mcu_fd_sc_mcu7t5v0__nand3_1_ZN_A3_gf180mcu_fd_ \
            #     sc_mcu7t5v0__and3_1_A3_Z_gf180mcu_fd_sc_mcu7t5v0__buf_1_I_Z
            d.run_pass("autoname")

        d.run_pass(
            "write_verilog",
            "-noattr",
            "-noexpr",
            "-nohex",
            "-nodec",
            "-defparam",
            output,
        )
        d.run_pass("write_json", f"{output}.json")

    run_strategy(d)

    if config["SYNTH_NO_FLAT"]:
        # Resynthesize, flattening
        d_flat = ys.Design()
        d_flat.add_blackbox_models(blackbox_models)

        shutil.copy(output, f"{output}.hierarchy.nl.v")
        d_flat.run_pass("read_verilog", "-sv", output)
        d_flat.run_pass(
            "synth",
            "-flatten",
            *(["-booth"] if config["SYNTH_MUL_BOOTH"] else []),
        )
        run_strategy(d_flat)


if __name__ == "__main__":
    synthesize()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from .flow import Flow
from .sequential import SequentialFlow
from ..steps import KLayout, OpenROAD, Magic


@Flow.factory.register()
class OpenInKLayout(SequentialFlow):
    """
    This 'flow' actually just has one step that opens the LEF/DEF from the
    initial state object in KLayout. Fancy that.

    Intended for use with run tags that have already been run with
    another flow, i.e.: ::

      openlane [...]
      openlane --last-run --flow OpenInKLayout [...]
    """

    name = "Opening in KLayout"
    Steps = [KLayout.OpenGUI]


@Flow.factory.register()
class OpenInOpenROAD(SequentialFlow):
    """
    This 'flow' actually just has one step that opens the ODB from
    the initial state object in OpenROAD.

    Intended for use with run tags that have already been run with
    another flow, i.e. ::

      openlane [...]
      openlane --last-run --flow OpenInOpenROAD [...]
    """

    name = "Opening in OpenROAD"

    Steps = [OpenROAD.OpenGUI]


@Flow.factory.register()
class OpenInMagic(SequentialFlow):
    """
    This 'flow' actually just has one step that opens the GDS or DEF from
    the initial state object in Magic.

    Intended for use with run tags that have already been run with
    another flow, i.e. ::

      openlane [...]
      openlane --last-run --flow OpenInMagic [...]
    """

    name = "Opening in Magic"

    Steps = [Magic.OpenGUI]



# Copyright 2024 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import io
import json
import fnmatch
import subprocess
from decimal import Decimal
from abc import abstractmethod
from typing import List, Literal, Optional, Set, Tuple

from .step import ViewsUpdate, MetricsUpdate, Step

from ..config import Variable
from ..state import State, DesignFormat
from ..logging import debug, verbose
from ..common import Path, get_script_dir, process_list_file

verilog_rtl_cfg_vars = [
    Variable(
        "VERILOG_FILES",
        List[Path],
        "The paths of the design's Verilog files.",
    ),
    Variable(
        "VERILOG_DEFINES",
        Optional[List[str]],
        "Preprocessor defines for input Verilog files.",
        deprecated_names=["SYNTH_DEFINES"],
    ),
    Variable(
        "VERILOG_POWER_DEFINE",
        str,
        "Specifies the name of the define used to guard power and ground connections in the input RTL.",
        deprecated_names=["SYNTH_USE_PG_PINS_DEFINES", "SYNTH_POWER_DEFINE"],
        default="USE_POWER_PINS",
    ),
    Variable(
        "VERILOG_INCLUDE_DIRS",
        Optional[List[str]],
        "Specifies the Verilog `include` directories.",
    ),
    Variable(
        "USE_SYNLIG",
        bool,
        "Use the Synlig plugin to process files, which has better SystemVerilog parsing capabilities but may not be compatible with all Yosys commands and attributes.",
        default=False,
    ),
    Variable(
        "SYNLIG_DEFER",
        bool,
        "Uses -defer flag when reading files the Synlig plugin, which may improve performance by reading each file separately, but is experimental.",
        default=False,
    ),
]

starts_with_whitespace = re.compile(r"^\s+.+$")

yosys_cell_rx = r"cell\s+\S+\s+\((\S+)\)"


def _check_any_tristate(
    cells: List[str],
    tristate_patterns: List[str],
):
    for cell in cells:
        for tristate_pattern in tristate_patterns:
            if fnmatch.fnmatch(cell, tristate_pattern):
                return True

    return False


def _parse_yosys_check(
    report: io.TextIOBase,
    tristate_patterns: Optional[List[str]] = None,
    tristate_okay: bool = False,
    elaborate_only: bool = False,
) -> int:
    verbose("Parsing synthesis checks…")
    errors_encountered: int = 0
    last_warning = None
    current_warning = None

    tristate_patterns = tristate_patterns or []

    for line in report:
        if line.startswith("Warning:") or line.startswith("Found and reported"):
            last_warning = current_warning
            current_warning = line
            if last_warning is None:
                continue

            cells = re.findall(yosys_cell_rx, last_warning)

            if elaborate_only and "but has no driver" in last_warning:
                debug("Ignoring undriven cell in elaborate-only mode:")
                debug(last_warning)
            elif tristate_okay and (
                ("tribuf" in last_warning)
                or _check_any_tristate(cells, tristate_patterns)
            ):
                debug("Ignoring tristate-related error:")
                debug(last_warning)
            else:
                debug("Encountered check error:")
                debug(last_warning)
                errors_encountered += 1
        elif (
            starts_with_whitespace.match(line) is not None
            and current_warning is not None
        ):
            current_warning += line
        else:
            pass
    return errors_encountered


verilog_rtl_cfg_vars = [
    Variable(
        "VERILOG_FILES",
        List[Path],
        "The paths of the design's Verilog files.",
    ),
    Variable(
        "VERILOG_DEFINES",
        Optional[List[str]],
        "Preprocessor defines for input Verilog files.",
        deprecated_names=["SYNTH_DEFINES"],
    ),
    Variable(
        "VERILOG_POWER_DEFINE",
        str,
        "Specifies the name of the define used to guard power and ground connections in the input RTL.",
        deprecated_names=["SYNTH_USE_PG_PINS_DEFINES", "SYNTH_POWER_DEFINE"],
        default="USE_POWER_PINS",
    ),
    Variable(
        "VERILOG_INCLUDE_DIRS",
        Optional[List[str]],
        "Specifies the Verilog `include` directories.",
    ),
    Variable(
        "SYNTH_PARAMETERS",
        Optional[List[str]],
        "Key-value pairs to be `chparam`ed in Yosys, in the format `key1=value1`.",
    ),
    Variable(
        "USE_SYNLIG",
        bool,
        "Use the Synlig plugin to process files, which has better SystemVerilog parsing capabilities but may not be compatible with all Yosys commands and attributes.",
        default=False,
    ),
    Variable(
        "SYNLIG_DEFER",
        bool,
        "Uses -defer flag when reading files the Synlig plugin, which may improve performance by reading each file separately, but is experimental.",
        default=False,
    ),
]


class PyosysStep(Step):
    config_vars = [
        Variable(
            "SYNTH_LATCH_MAP",
            Optional[Path],
            "A path to a file containing the latch mapping for Yosys.",
            pdk=True,
        ),
        Variable(
            "SYNTH_TRISTATE_MAP",
            Optional[Path],
            "A path to a file containing the tri-state buffer mapping for Yosys.",
            deprecated_names=["TRISTATE_BUFFER_MAP"],
            pdk=True,
        ),
        Variable(
            "SYNTH_CSA_MAP",
            Optional[Path],
            "A path to a file containing the carry-select adder mapping for Yosys.",
            deprecated_names=["CARRY_SELECT_ADDER_MAP"],
            pdk=True,
        ),
        Variable(
            "SYNTH_RCA_MAP",
            Optional[Path],
            "A path to a file containing the ripple-carry adder mapping for Yosys.",
            deprecated_names=["RIPPLE_CARRY_ADDER_MAP"],
            pdk=True,
        ),
        Variable(
            "SYNTH_FA_MAP",
            Optional[Path],
            "A path to a file containing the full adder mapping for Yosys.",
            deprecated_names=["FULL_ADDER_MAP"],
            pdk=True,
        ),
        Variable(
            "SYNTH_MUX_MAP",
            Optional[Path],
            "A path to a file containing the mux mapping for Yosys.",
            pdk=True,
        ),
        Variable(
            "SYNTH_MUX4_MAP",
            Optional[Path],
            "A path to a file containing the mux4 mapping for Yosys.",
            pdk=True,
        ),
        Variable(
            "USE_LIGHTER",
            bool,
            "Activates Lighter, an experimental plugin that attempts to optimize clock-gated flip-flops.",
            default=False,
        ),
        Variable(
            "LIGHTER_DFF_MAP",
            Optional[Path],
            "An override to the custom DFF map file provided for the given SCL by Lighter.",
        ),
        Variable(
            "YOSYS_LOG_LEVEL",
            Literal["ALL", "WARNING", "ERROR"],
            "Which log level for Yosys. At WARNING or higher, the initialization splash is also disabled.",
            default="ALL",
        ),
    ]

    @abstractmethod
    def get_script_path(self) -> str:
        pass

    def get_command(self, state_in: State) -> List[str]:
        script_path = self.get_script_path()
        cmd = ["yosys", "-y", script_path]
        if self.config["YOSYS_LOG_LEVEL"] != "ALL":
            cmd += ["-Q"]
        if self.config["YOSYS_LOG_LEVEL"] == "WARNING":
            cmd += ["-q"]
        elif self.config["YOSYS_LOG_LEVEL"] == "ERROR":
            cmd += ["-qq"]
        cmd += ["--"]
        cmd += ["--config-in", os.path.join(self.step_dir, "config.json")]
        return cmd

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        cmd = self.get_command(state_in)
        subprocess_result = super().run_subprocess(cmd, **kwargs)
        return {}, subprocess_result["generated_metrics"]


class VerilogStep(PyosysStep):
    power_defines: bool = False

    config_vars = PyosysStep.config_vars + verilog_rtl_cfg_vars

    def get_command(self, state_in: State) -> List[str]:
        cmd = super().get_command(state_in)

        blackbox_models = []
        scl_lib_list = self.toolbox.filter_views(self.config, self.config["LIB"])
        if self.power_defines and self.config["CELL_VERILOG_MODELS"] is not None:
            blackbox_models.extend(
                [
                    self.toolbox.create_blackbox_model(
                        frozenset(self.config["CELL_VERILOG_MODELS"]),
                        frozenset(["USE_POWER_PINS"]),
                    )
                ]
            )
        else:
            blackbox_models.extend(str(f) for f in scl_lib_list)

        # Priorities from higher to lower
        format_list = (
            [
                DesignFormat.VERILOG_HEADER,
                DesignFormat.POWERED_NETLIST,
                DesignFormat.NETLIST,
                DesignFormat.LIB,
            ]
            if self.power_defines
            else [
                DesignFormat.VERILOG_HEADER,
                DesignFormat.NETLIST,
                DesignFormat.POWERED_NETLIST,
                DesignFormat.LIB,
            ]
        )
        for view, _ in self.toolbox.get_macro_views_by_priority(
            self.config, format_list
        ):
            blackbox_models.append(str(view))

        if libs := self.config.get("EXTRA_LIBS"):
            blackbox_models.extend(str(f) for f in libs)
        if models := self.config.get("EXTRA_VERILOG_MODELS"):
            blackbox_models.extend(str(f) for f in models)

        excluded_cells: Set[str] = set(self.config["EXTRA_EXCLUDED_CELLS"] or [])
        excluded_cells.update(
            process_list_file(self.config["SYNTH_EXCLUDED_CELL_FILE"])
        )
        excluded_cells.update(process_list_file(self.config["PNR_EXCLUDED_CELL_FILE"]))

        libs_synth = self.toolbox.remove_cells_from_lib(
            frozenset([str(lib) for lib in scl_lib_list]),
            excluded_cells=frozenset(excluded_cells),
        )
        extra_path = os.path.join(self.step_dir, "extra.json")
        with open(extra_path, "w") as f:
            json.dump({"blackbox_models": blackbox_models, "libs_synth": libs_synth}, f)
        cmd.extend(["--extra-in", extra_path])
        return cmd


@Step.factory.register()
class JsonHeader(VerilogStep):
    id = "Yosys.JsonHeader"
    name = "Generate JSON Header"
    long_name = "Generate JSON Header"

    inputs = []
    outputs = [DesignFormat.JSON_HEADER]

    config_vars = PyosysStep.config_vars + verilog_rtl_cfg_vars

    power_defines = True

    def get_script_path(self) -> str:
        return os.path.join(get_script_dir(), "pyosys", "json_header.py")

    def get_command(self, state_in: State) -> List[str]:
        out_file = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.JSON_HEADER.value.extension}",
        )
        return super().get_command(state_in) + ["--output", out_file]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        out_file = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.JSON_HEADER.value.extension}",
        )
        views_updates, metrics_updates = super().run(state_in, **kwargs)
        views_updates[DesignFormat.JSON_HEADER] = Path(out_file)
        return views_updates, metrics_updates


class SynthesisCommon(VerilogStep):
    inputs = []  # The input RTL is part of the configuration
    outputs = [DesignFormat.NETLIST]

    config_vars = PyosysStep.config_vars + [
        Variable(
            "SYNTH_CHECKS_ALLOW_TRISTATE",
            bool,
            "Ignore multiple-driver warnings if they are connected to tri-state buffers on a best-effort basis.",
            default=True,
        ),
        Variable(
            "SYNTH_AUTONAME",
            bool,
            "Generates names for netlist instances. This results in instance names that can be extremely long, but are more human-readable.",
            default=False,
        ),
        Variable(
            "SYNTH_STRATEGY",
            Literal[
                "AREA 0",
                "AREA 1",
                "AREA 2",
                "AREA 3",
                "DELAY 0",
                "DELAY 1",
                "DELAY 2",
                "DELAY 3",
                "DELAY 4",
            ],
            "Strategies for abc logic synthesis and technology mapping. AREA strategies usually result in a more compact design, while DELAY strategies usually result in a design that runs at a higher frequency. Please note that there is no way to know which strategy is the best before trying them.",
            default="AREA 0",
        ),
        Variable(
            "SYNTH_ABC_BUFFERING",
            bool,
            "Enables `abc` cell buffering.",
            default=False,
            deprecated_names=["SYNTH_BUFFERING"],
        ),
        Variable(
            "SYNTH_ABC_LEGACY_REFACTOR",
            bool,
            "Replaces the ABC command `drf -l` with `refactor` which matches older versions of OpenLane but is more unstable.",
            default=False,
        ),
        Variable(
            "SYNTH_ABC_LEGACY_REWRITE",
            bool,
            "Replaces the ABC command `drw -l` with `rewrite` which matches older versions of OpenLane but is more unstable.",
            default=False,
        ),
        Variable(
            "SYNTH_ABC_DFF",
            bool,
            "Passes D-flipflop cells through ABC for optimization (which can for example, eliminate identical flip-flops).",
            default=False,
        ),
        Variable(
            "SYNTH_ABC_USE_MFS3",
            bool,
            "Experimental: attempts a SAT-based remapping in all area and delay strategies before 'retime', which may improve PPA results.",
            default=False,
        ),
        Variable(
            "SYNTH_ABC_AREA_USE_NF",
            bool,
            "Experimental: uses the &nf delay-based mapper with a very high value instead of the amap area mapper, which may be better in some scenarios at recovering area.",
            default=False,
        ),
        Variable(
            "SYNTH_DIRECT_WIRE_BUFFERING",
            bool,
            "Enables inserting buffer cells for directly connected wires.",
            default=True,
            deprecated_names=["SYNTH_BUFFER_DIRECT_WIRES"],
        ),
        Variable(
            "SYNTH_SPLITNETS",
            bool,
            "Splits multi-bit nets into single-bit nets. Easier to trace but may not be supported by all tools.",
            default=True,
        ),
        Variable(
            "SYNTH_SIZING",
            bool,
            "Enables `abc` cell sizing (instead of buffering).",
            default=False,
        ),
        Variable(
            "SYNTH_NO_FLAT",
            bool,
            "A flag that disables flattening the hierarchy during synthesis, only flattening it after synthesis, mapping and optimizations.",
            default=False,
        ),
        Variable(
            "SYNTH_SHARE_RESOURCES",
            bool,
            "A flag that enables yosys to reduce the number of cells by determining shareable resources and merging them.",
            default=True,
        ),
        Variable(
            "SYNTH_ADDER_TYPE",
            Literal["YOSYS", "FA", "RCA", "CSA"],
            "Adder type to which the $add and $sub operators are mapped to.  Possible values are `YOSYS/FA/RCA/CSA`; where `YOSYS` refers to using Yosys internal adder definition, `FA` refers to full-adder structure, `RCA` refers to ripple carry adder structure, and `CSA` refers to carry select adder.",
            default="YOSYS",
        ),
        Variable(
            "SYNTH_EXTRA_MAPPING_FILE",
            Optional[Path],
            "Points to an extra techmap file for yosys that runs right after yosys `synth` before generic techmap.",
        ),
        Variable(
            "SYNTH_ELABORATE_ONLY",
            bool,
            '"Elaborate" the design only without attempting any logic mapping. Useful when dealing with structural Verilog netlists.',
            default=False,
        ),
        Variable(
            "SYNTH_ELABORATE_FLATTEN",
            bool,
            "If `SYNTH_ELABORATE_ONLY` is specified, this variable controls whether or not the top level should be flattened.",
            default=True,
            deprecated_names=["SYNTH_FLAT_TOP"],
        ),
        Variable(
            "SYNTH_MUL_BOOTH",
            bool,
            "Runs the booth pass as part of synthesis: See https://yosyshq.readthedocs.io/projects/yosys/en/latest/cmd/booth.html",
            default=False,
        ),
        # Variable(
        #     "SYNTH_SDC_FILE",
        #     Optional[Path],
        #     "Specifies the SDC file read during all Synthesis steps",
        # ),
    ]

    def get_script_path(self) -> str:
        return os.path.join(get_script_dir(), "pyosys", "synthesize.py")

    def get_command(self, state_in: State) -> List[str]:
        out_file = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.NETLIST.value.extension}",
        )
        cmd = super().get_command(state_in)
        if self.config["USE_LIGHTER"]:
            lighter_dff_map = self.config["LIGHTER_DFF_MAP"]
            if lighter_dff_map is None:
                scl = self.config["STD_CELL_LIBRARY"]
                try:
                    raw = subprocess.check_output(
                        ["lighter_files", scl], encoding="utf8"
                    )
                    files = raw.strip().splitlines()
                    lighter_dff_map = Path(files[0])
                except FileNotFoundError:
                    self.warn(
                        "Lighter not found or not set up with OpenLane: If you're using a manual Lighter install, try setting LIGHTER_DFF_MAP explicitly."
                    )
                except subprocess.CalledProcessError:
                    self.warn(f"{scl} not supported by Lighter.")
            cmd.extend(["--lighter-dff-map", lighter_dff_map])
        return cmd + ["--output", out_file]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        out_file = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.NETLIST.value.extension}",
        )

        view_updates, metric_updates = super().run(state_in, **kwargs)

        stats_file = os.path.join(self.step_dir, "reports", "stat.json")
        stats_str = open(stats_file).read()
        stats = json.loads(stats_str, parse_float=Decimal)

        metric_updates["design__instance__count"] = stats["design"]["num_cells"]
        if chip_area := stats["design"].get("area"):  # needs nonzero area
            metric_updates["design__instance__area"] = chip_area

        cells = stats["design"]["num_cells_by_type"]
        safe = ["$assert"]
        unmapped_cells = [
            cells[y] for y in cells.keys() if y not in safe and y.startswith("$")
        ]
        metric_updates["design__instance_unmapped__count"] = sum(unmapped_cells)

        check_error_count_file = os.path.join(
            self.step_dir, "reports", "pre_synth_chk.rpt"
        )
        metric_updates["synthesis__check_error__count"] = 0
        if os.path.exists(check_error_count_file):
            metric_updates["synthesis__check_error__count"] = _parse_yosys_check(
                open(check_error_count_file),
                self.config["TRISTATE_CELLS"],
                self.config["SYNTH_CHECKS_ALLOW_TRISTATE"],
                self.config["SYNTH_ELABORATE_ONLY"],
            )

        view_updates[DesignFormat.NETLIST] = Path(out_file)

        return view_updates, metric_updates


@Step.factory.register()
class Synthesis(SynthesisCommon):
    """
    Performs synthesis and technology mapping on Verilog RTL files
    using Yosys and ABC, emitting a netlist.

    Some metrics will also be extracted and updated, namely:

    * ``design__instance__count``
    * ``design__instance_unmapped__count``
    * ``design__instance__area``
    """

    id = "Yosys.Synthesis"
    name = "Synthesis"

    config_vars = SynthesisCommon.config_vars + verilog_rtl_cfg_vars


@Step.factory.register()
class VHDLSynthesis(SynthesisCommon):
    """
    Performs synthesis and technology mapping on VHDL files
    using Yosys, GHDL and ABC, emitting a netlist.

    Some metrics will also be extracted and updated, namely:

    * ``design__instance__count``
    * ``design__instance_unmapped__count``
    * ``design__instance__area``
    """

    id = "Yosys.VHDLSynthesis"
    name = "Synthesis (VHDL)"

    config_vars = SynthesisCommon.config_vars + [
        Variable(
            "VHDL_FILES",
            List[Path],
            "The paths of the design's VHDL files.",
        ),
    ]



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import sys
import json
import shutil
from decimal import Decimal
from typing import Callable, List, Mapping, Tuple, Union, Optional, Dict, Any

from .design_format import (
    DesignFormat,
    DesignFormatObject,
)

from ..common import (
    Path,
    GenericImmutableDict,
    mkdirp,
    copy_recursive,
)
from ..logging import info


class InvalidState(RuntimeError):
    pass


StateElement = Union[Path, List[Path], Dict[str, Union[Path, List[Path]]], None]


class State(GenericImmutableDict[str, StateElement]):
    """
    Basically, a dictionary from :class:`DesignFormat`\\s and values
    of (nested dictionaries of) :class:`openlane.common.Path`\\.

    The state is the only thing that can be altered by steps other than the
    filesystem.

    States are **immutable**. To construct a new state with some modifications,
    you may do so as follows::

        state_b = State(
            copying=state_a,
            overrides={
                …
            },
            metrics=GenericImmutableDict(copying=state_a.metrics, overrides={
                …
            })
        )

    Though in the majority of cases, you do not have to construct States on your
    own: after executing a Step, you only return your deltas and then the Flow
    is responsible for the creation of a new Step object.

    :param copying: A mutable or immutable mapping to use as the starting
        value for this State.
    :param overrides: A mutable or immutable mapping to override the starting
        values with.
    :param metrics: A dictionary that carries statistics about the design: area,
        wire length, et cetera, but also miscellaneous data, for example, whether
        it passed a certain check or not.
    """

    def __init__(
        self,
        copying: Optional[
            Union[Mapping[str, StateElement], Mapping[DesignFormat, StateElement]]
        ] = None,
        *args,
        overrides: Optional[
            Union[Mapping[str, StateElement], Mapping[DesignFormat, StateElement]]
        ] = None,
        metrics: Optional[Mapping[str, Any]] = None,
        **kwargs,
    ) -> None:
        copying_resolved: Dict[str, StateElement] = {}
        if c_mapping := copying:
            for key, value in c_mapping.items():
                if isinstance(key, DesignFormat):
                    copying_resolved[key.value.id] = value
                else:
                    copying_resolved[key] = value

        for format in DesignFormat:
            assert isinstance(format.value, DesignFormatObject)  # type checker shut up
            if format.value.id not in copying_resolved:
                copying_resolved[format.value.id] = None

        overrides_resolved = {}
        if o_mapping := overrides:
            for k, value in o_mapping.items():
                if isinstance(k, DesignFormat):
                    assert isinstance(
                        k.value, DesignFormatObject
                    )  # type checker shut up
                    k = k.value.id
                overrides_resolved[k] = value

        self.metrics = GenericImmutableDict(metrics or {})

        super().__init__(
            copying_resolved,
            *args,
            overrides=overrides_resolved,
            **kwargs,
        )

    def __getitem__(self, key: Union[DesignFormat, str]) -> StateElement:
        if isinstance(key, DesignFormat):
            id: str = key.value.id
            key = id
        return super().__getitem__(key)

    def __setitem__(self, key: Union[DesignFormat, str], item: StateElement):
        if isinstance(key, DesignFormat):
            id: str = key.value.id
            key = id
        return super().__setitem__(key, item)

    def __delitem__(self, key: Union[DesignFormat, str]):
        if isinstance(key, DesignFormat):
            id: str = key.value.id
            key = id
        return super().__delitem__(key)

    def to_raw_dict(self, metrics: bool = True) -> Dict[str, Any]:
        final = super().to_raw_dict()
        if metrics:
            final["metrics"] = self.metrics.to_raw_dict()
        return final

    def copy(self: "State") -> "State":
        metrics: GenericImmutableDict[str, Any] = GenericImmutableDict(
            copy_recursive(self.metrics)
        )
        new = State(self, metrics=metrics)
        return new

    def _walk(
        self,
        views: Union[Dict, "State"],
        save_directory: Union[str, os.PathLike],
        visit: Callable[[str, StateElement, str, str, int], StateElement],
        key_path: str = "",
        depth: int = 0,
        top_key: Optional[str] = None,
    ):
        for key, value in views.items():
            current_top_key = top_key
            if current_top_key is None:
                current_top_key = key
            current_folder = key.strip("_*")
            if df := DesignFormat.by_id(key):
                # For type-checker: all guaranteed to be DesignFormatObjects
                assert isinstance(df.value, DesignFormatObject)
                current_folder = df.value.folder

            target_dir = os.path.join(save_directory, current_folder)

            current_key_path = f"{key_path}.{key}"
            visit(current_key_path, value, current_top_key, target_dir, depth)
            if isinstance(value, dict):
                self._walk(
                    value,
                    target_dir,
                    visit,
                    current_key_path,
                    depth + 1,
                    current_top_key,
                )
            if isinstance(value, list):
                for i, element in enumerate(value):
                    element_key_path = f"{current_key_path}[{i}]"
                    visit(
                        element_key_path,
                        element,
                        current_top_key,
                        target_dir,
                        depth + 1,
                    )

    def save_snapshot(self, path: Union[str, os.PathLike]):
        """
        Validates the current state then saves all views to a folder by
        design format, including the metrics.

        :param path: The folder that would contain other folders.
        """

        def visitor(key, value, top_key, save_directory, depth):
            if not isinstance(value, Path):
                return
            mkdirp(save_directory)
            target_path = os.path.join(save_directory, os.path.basename(value))
            shutil.copyfile(value, target_path, follow_symlinks=True)

        self.validate()
        info(f"Saving views to '{os.path.abspath(path)}'…")
        mkdirp(path)
        self._walk(self, path, visitor)
        metrics_csv_path = os.path.join(path, "metrics.csv")
        with open(metrics_csv_path, "w", encoding="utf8") as f:
            f.write("Metric,Value\n")
            for metric in self.metrics:
                f.write(f"{metric},{self.metrics[metric]}\n")

        metrics_json_path = os.path.join(path, "metrics.json")
        with open(metrics_json_path, "w", encoding="utf8") as f:
            f.write(self.metrics.dumps())

    def validate(self):
        """
        Ensures that all paths exist in a State.
        """

        def visitor(key, value, top_key, _, depth):
            if depth == 0 and DesignFormat.by_id(top_key) is None:
                raise InvalidState(
                    f"Key '{top_key}' does not match a known design format."
                )
            if value is None:
                return
            if not (
                isinstance(value, Path)
                or isinstance(value, dict)
                or isinstance(value, list)
            ):
                raise InvalidState(
                    f"Value at '{key}' is not a Path nor a dictionary/list of Paths: '{value}'."
                )

        self._walk(self.to_raw_dict(metrics=False), "", visit=visitor)

    @classmethod
    def __loads_recursive(
        Self,
        views: Dict,
        validate_path: bool = True,
        key_path: str = "",
    ) -> dict:
        target: dict = {}
        for key, value in views.items():
            current_key_path = f"{key_path}.{key}"
            if value is None:
                target[key] = value
                continue

            if isinstance(value, dict):
                target[key] = Self.__loads_recursive(
                    value,
                    validate_path,
                    key_path=current_key_path,
                )
            else:
                if validate_path and not os.path.exists(value):
                    raise ValueError(
                        f"Provided path '{value}' to design format '{current_key_path}' does not exist."
                    )
                target[key] = Path(value)
        return target

    @classmethod
    def loads(Self, json_in: str, validate_path: bool = True) -> "State":
        try:
            raw = json.loads(json_in, parse_float=Decimal)
        except json.JSONDecodeError as e:
            raise InvalidState(f"Invalid JSON string provided for state: {e}")

        if not isinstance(raw, dict):
            raise InvalidState("Failed to load state: JSON result is not a dictionary")

        metrics = raw.get("metrics")
        if metrics is not None:
            del raw["metrics"]

        views = Self.__loads_recursive(raw, validate_path)
        state = Self(views, metrics=metrics)

        return state

    def __mapping_to_html_rec(
        self,
        mapping: Mapping[str, Any],
        header_optional: Optional[Tuple[str, str]] = None,
    ):
        result = """
        <table style="grid-column-start: 1; grid-column-end: 2; ">
        """
        if header := header_optional:
            key_h, value_h = header
            result += f"""
                <tr>
                    <th style="text-align: left;">{key_h}</th>
                    <th style="text-align: left;">{value_h}</th>
                </tr>
            """

        for id, value in mapping.items():
            if value is None:
                continue

            key_content = id
            if format := DesignFormat.by_id(id):
                key_content = format.value.id

            value_content = str(value)
            if isinstance(value, Mapping):
                value_content = self.__mapping_to_html_rec(value)
            elif isinstance(value, Path):
                value_rel = os.path.relpath(value, ".")

                value_content = f'<a href="{value_rel}">{value_rel}</a>'
                if "google.colab" in sys.modules:
                    # Can't link in colab
                    value_content = value_rel

            result += f"""
                <tr>
                    <td style="text-align: left;">{key_content}</td>
                    <td style="text-align: left;">{value_content}</td>
                </tr>
            """

        result += """
        </table>
        """
        return result

    def _repr_html_(self) -> str:
        return (
            '<div style="display: grid; grid-auto-columns: minmax(0, 1fr); grid-auto-rows: minmax(0, 1fr); grid-auto-flow: column;">'
            + self.__mapping_to_html_rec(
                self.to_raw_dict(metrics=False),
                ("Format", "Path"),
            )
            + "</div>"
        )



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import site
import shlex
import shutil
import subprocess
from os.path import abspath
from base64 import b64encode
from typing import Any, Dict, Optional, List, Sequence, Tuple, Union

from .step import ViewsUpdate, MetricsUpdate, Step, StepError, StepException

from ..config import Variable
from ..logging import info
from ..state import DesignFormat, State
from ..common import Path, get_script_dir, mkdirp, _get_process_limit


class KLayoutStep(Step):
    config_vars = [
        Variable(
            "KLAYOUT_TECH",
            Path,
            "A path to the KLayout layer technology (.lyt) file.",
            pdk=True,
        ),
        Variable(
            "KLAYOUT_PROPERTIES",
            Path,
            "A path to the KLayout layer properties (.lyp) file.",
            pdk=True,
        ),
        Variable(
            "KLAYOUT_DEF_LAYER_MAP",
            Path,
            "A path to the KLayout LEF/DEF layer mapping (.map) file.",
            pdk=True,
        ),
    ]

    def run_pya_script(
        self,
        cmd: Sequence[Union[str, os.PathLike]],
        log_to: Optional[Union[str, os.PathLike]] = None,
        silent: bool = False,
        report_dir: Optional[Union[str, os.PathLike]] = None,
        env: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        env = env or os.environ.copy()
        # Pass site packages
        python_path_elements = site.getsitepackages() + sys.path
        if current_pythonpath := env.get("PYTHONPATH"):
            python_path_elements.append(current_pythonpath)

        env["PYTHONPATH"] = ":".join(python_path_elements)
        return super().run_subprocess(cmd, log_to, silent, report_dir, env, **kwargs)

    def get_cli_args(
        self,
        *,
        layer_info: bool = True,
        include_lefs: bool = False,
        include_gds: bool = False,
    ) -> List[str]:
        result = []
        if layer_info:
            lyp = abspath(self.config["KLAYOUT_PROPERTIES"])
            lyt = abspath(self.config["KLAYOUT_TECH"])
            lym = abspath(self.config["KLAYOUT_DEF_LAYER_MAP"])
            if None in [lyp, lyt, lym]:
                raise StepError(
                    "Cannot open design in KLayout as the PDK does not appear to support KLayout."
                )
            result += ["--lyp", lyp, "--lyt", lyt, "--lym", lym]

        if include_lefs:
            tech_lefs = self.toolbox.filter_views(self.config, self.config["TECH_LEFS"])
            if len(tech_lefs) != 1:
                raise StepException(
                    "Misconfigured SCL: 'TECH_LEFS' must return exactly one Tech LEF for its default timing corner."
                )

            lef_args = [
                "--input-lef",
                abspath(tech_lefs[0]),
            ]

            for lef in self.config["CELL_LEFS"]:
                lef_args.append("--input-lef")
                lef_args.append(abspath(lef))

            macro_lefs = self.toolbox.get_macro_views(self.config, DesignFormat.LEF)
            for lef in macro_lefs:
                lef_args.append("--input-lef")
                lef_args.append(abspath(lef))

            if extra_lefs := self.config["EXTRA_LEFS"]:
                for lef in extra_lefs:
                    lef_args.append("--input-lef")
                    lef_args.append(abspath(lef))

            result += lef_args

        if include_gds:
            gds_args = []
            for gds in self.config["CELL_GDS"]:
                gds_args.append("--with-gds-file")
                gds_args.append(gds)
            for gds in self.toolbox.get_macro_views(self.config, DesignFormat.GDS):
                gds_args.append("--with-gds-file")
                gds_args.append(gds)
            if extra_gds := self.config["EXTRA_GDS_FILES"]:
                for gds in extra_gds:
                    gds_args.append("--with-gds-file")
                    gds_args.append(gds)
            result += gds_args

        return result


@Step.factory.register()
class Render(KLayoutStep):
    """
    Renders a PNG of the layout using KLayout.

    DEF is required as an input, but if a GDS-II view
    exists in the input state, it will be used instead.
    """

    id = "KLayout.Render"
    name = "Render Image (w/ KLayout)"

    inputs = [DesignFormat.DEF]
    outputs = []

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        input_view = state_in[DesignFormat.DEF]
        if gds := state_in[DesignFormat.GDS]:
            input_view = gds

        assert isinstance(input_view, Path)

        self.run_pya_script(
            [
                sys.executable,
                os.path.join(get_script_dir(), "klayout", "render.py"),
                abspath(input_view),
                "--output",
                abspath(os.path.join(self.step_dir, "out.png")),
            ]
            + self.get_cli_args(include_lefs=True),
            silent=True,
        )

        return {}, {}


@Step.factory.register()
class StreamOut(KLayoutStep):
    """
    Converts DEF views into GDSII streams using KLayout.

    The PDK must support KLayout for this step to work, otherwise
    it will be skipped.

    If ``PRIMARY_GDSII_STREAMOUT_TOOL`` is set to ``"klayout"``, both GDS and KLAYOUT_GDS
    will be updated, and if set to another tool, only ``KLAYOUT_GDS`` will be
    updated.
    """

    id = "KLayout.StreamOut"
    name = "GDSII Stream Out (KLayout)"

    inputs = [DesignFormat.DEF]
    outputs = [DesignFormat.GDS, DesignFormat.KLAYOUT_GDS]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        views_updates: ViewsUpdate = {}

        klayout_gds_out = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.KLAYOUT_GDS.value.extension}",
        )
        kwargs, env = self.extract_env(kwargs)

        self.run_pya_script(
            [
                sys.executable,
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "stream_out.py",
                ),
                state_in[DesignFormat.DEF.value.id],
                "--output",
                abspath(klayout_gds_out),
                "--top",
                self.config["DESIGN_NAME"],
            ]
            + self.get_cli_args(include_lefs=True, include_gds=True),
            env=env,
        )

        views_updates[DesignFormat.KLAYOUT_GDS] = Path(klayout_gds_out)

        if self.config["PRIMARY_GDSII_STREAMOUT_TOOL"] == "klayout":
            gds_path = os.path.join(self.step_dir, f"{self.config['DESIGN_NAME']}.gds")
            shutil.copy(klayout_gds_out, gds_path)
            views_updates[DesignFormat.GDS] = Path(gds_path)

        return views_updates, {}

    def layout_preview(self) -> Optional[str]:
        if self.state_out is None:
            return None
        assert self.toolbox is not None

        if image := self.toolbox.render_png(self.config, self.state_out):
            image_encoded = b64encode(image).decode("utf8")
            return f'<img src="data:image/png;base64,{image_encoded}" />'

        return None


@Step.factory.register()
class XOR(KLayoutStep):
    """
    Performs an XOR operation on the Magic and KLayout GDS views. The idea is:
    if there's any difference between the GDSII streams between the two tools,
    one of them have it wrong and that may lead to ambiguity.
    """

    id = "KLayout.XOR"
    name = "KLayout vs. Magic XOR"

    inputs = [
        DesignFormat.MAG_GDS,
        DesignFormat.KLAYOUT_GDS,
    ]
    outputs = []

    config_vars = KLayoutStep.config_vars + [
        Variable(
            "KLAYOUT_XOR_THREADS",
            Optional[int],
            "Specifies number of threads used in the KLayout XOR check. If unset, this will be equal to your machine's thread count.",
        ),
        Variable(
            "KLAYOUT_XOR_IGNORE_LAYERS",
            Optional[List[str]],
            "KLayout layers to ignore during XOR operations.",
            pdk=True,
        ),
        Variable(
            "KLAYOUT_XOR_TILE_SIZE",
            Optional[int],
            "A tile size for the XOR process in µm.",
            pdk=True,
        ),
    ]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        ignored = ""
        if ignore_list := self.config["KLAYOUT_XOR_IGNORE_LAYERS"]:
            ignored = ";".join(ignore_list)

        layout_a = state_in[DesignFormat.MAG_GDS]
        if layout_a is None:
            self.warn("No Magic stream-out has been performed. Skipping XOR process…")
            return {}, {}
        layout_b = state_in[DesignFormat.KLAYOUT_GDS]
        if layout_b is None:
            self.warn("No KLayout stream-out has been performed. Skipping XOR process…")
            return {}, {}

        assert isinstance(layout_a, Path)
        assert isinstance(layout_b, Path)

        kwargs, env = self.extract_env(kwargs)

        tile_size_options = []
        if tile_size := self.config["KLAYOUT_XOR_TILE_SIZE"]:
            tile_size_options += ["--tile-size", str(tile_size)]

        thread_count = self.config["KLAYOUT_XOR_THREADS"] or _get_process_limit()
        info(f"Running XOR with {thread_count} threads…")

        subprocess_result = self.run_subprocess(
            [
                "ruby",
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "xor.drc",
                ),
                "--output",
                abspath(os.path.join(self.step_dir, "xor.xml")),
                "--top",
                self.config["DESIGN_NAME"],
                "--threads",
                thread_count,
                "--ignore",
                ignored,
                abspath(layout_a),
                abspath(layout_b),
            ]
            + tile_size_options,
            env=env,
        )

        return {}, subprocess_result["generated_metrics"]


@Step.factory.register()
class DRC(KLayoutStep):
    id = "KLayout.DRC"
    name = "Design Rule Check (KLayout)"

    inputs = [
        DesignFormat.GDS,
    ]
    outputs = []

    config_vars = KLayoutStep.config_vars + [
        Variable(
            "KLAYOUT_DRC_RUNSET",
            Optional[Path],
            "A path to KLayout DRC runset.",
            pdk=True,
            deprecated_names=["KLAYOUT_DRC_TECH_SCRIPT"],
        ),
        Variable(
            "KLAYOUT_DRC_OPTIONS",
            Optional[Dict[str, Union[bool, int]]],
            "Options passed directly to the KLayout DRC runset. They vary from one PDK to another.",
            pdk=True,
        ),
        Variable(
            "KLAYOUT_DRC_THREADS",
            Optional[int],
            "Specifies the number of threads to be used in KLayout DRC"
            + "If unset, this will be equal to your machine's thread count.",
        ),
    ]

    def run_sky130(self, state_in: State, **kwargs) -> MetricsUpdate:
        kwargs, env = self.extract_env(kwargs)
        reports_dir = os.path.join(self.step_dir, "reports")
        mkdirp(reports_dir)
        drc_script_path = self.config["KLAYOUT_DRC_RUNSET"]
        xml_report = os.path.join(reports_dir, "drc_violations.klayout.xml")
        json_report = os.path.join(reports_dir, "drc_violations.klayout.json")
        feol = str(self.config["KLAYOUT_DRC_OPTIONS"]["feol"]).lower()
        beol = str(self.config["KLAYOUT_DRC_OPTIONS"]["beol"]).lower()
        floating_metal = str(
            self.config["KLAYOUT_DRC_OPTIONS"]["floating_metal"]
        ).lower()
        offgrid = str(self.config["KLAYOUT_DRC_OPTIONS"]["offgrid"]).lower()
        seal = str(self.config["KLAYOUT_DRC_OPTIONS"]["seal"]).lower()
        threads = self.config["KLAYOUT_DRC_THREADS"] or _get_process_limit()
        info(f"Running KLayout DRC with {threads} threads…")

        input_view = state_in[DesignFormat.GDS]
        assert isinstance(input_view, Path)

        # Not pya script - DRC script is not part of OpenLane
        self.run_subprocess(
            [
                "klayout",
                "-b",
                "-zz",
                "-r",
                drc_script_path,
                "-rd",
                f"input={abspath(input_view)}",
                "-rd",
                f"topcell={self.config['DESIGN_NAME']}",
                "-rd",
                f"report={abspath(xml_report)}",
                "-rd",
                f"feol={feol}",
                "-rd",
                f"beol={beol}",
                "-rd",
                f"floating_metal={floating_metal}",
                "-rd",
                f"offgrid={offgrid}",
                "-rd",
                f"seal={seal}",
                "-rd",
                f"threads={threads}",
            ],
            env=env,
        )

        subprocess_result = self.run_pya_script(
            [
                "python3",
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "xml_drc_report_to_json.py",
                ),
                f"--xml-file={abspath(xml_report)}",
                f"--json-file={abspath(json_report)}",
            ],
            env=env,
            log_to=os.path.join(self.step_dir, "xml_drc_report_to_json.log"),
        )
        return subprocess_result["generated_metrics"]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        metrics_updates: MetricsUpdate = {}
        if self.config["PDK"] in ["sky130A", "sky130B"]:
            metrics_updates = self.run_sky130(state_in, **kwargs)
        else:
            self.warn(
                f"KLayout DRC is not supported for the {self.config['PDK']} PDK. This step will be skipped."
            )

        return {}, metrics_updates


@Step.factory.register()
class OpenGUI(KLayoutStep):
    """
    Opens the DEF view in the KLayout GUI, with layers loaded and mapped
    properly. Useful to inspect ``.klayout.xml`` database files and the like.
    """

    id = "KLayout.OpenGUI"
    name = "Open In GUI"

    inputs = [DesignFormat.DEF]
    outputs = []

    config_vars = KLayoutStep.config_vars + [
        Variable(
            "KLAYOUT_EDITOR_MODE",
            bool,
            "Whether to run the KLayout GUI in editor mode or in viewer mode.",
            default=False,
        ),
        Variable(
            "KLAYOUT_GUI_USE_GDS",
            bool,
            "Whether to prioritize GDS (if found) when running this step.",
            default=True,
            deprecated_names=["KLAYOUT_PRIORITIZE_GDS"],
        ),
    ]

    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        kwargs, env = self.extract_env(kwargs)
        mode_args = []
        if self.config["KLAYOUT_EDITOR_MODE"]:
            mode_args.append("--editor")

        layout = state_in[DesignFormat.DEF]
        if self.config["KLAYOUT_GUI_USE_GDS"]:
            if gds := state_in[DesignFormat.GDS]:
                layout = gds
        assert isinstance(layout, Path)

        env["KLAYOUT_ARGV"] = shlex.join(
            [
                abspath(layout),
            ]
            + self.get_cli_args(include_lefs=True)
        )

        cmd = (
            [
                shutil.which("klayout") or "klayout",
            ]
            + mode_args
            + [
                "-rm",
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "open_design.py",
                ),
            ]
        )

        # Not run_subprocess- need stdin, stdout, stderr to be accessible to the
        # user normally
        subprocess.check_call(
            cmd,
            env=env,
            cwd=self.step_dir,
        )

        return {}, {}



# Copyright 2021-2022 Efabless Corporation
# Copyright 2022 Arman Avetisyan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odb

import os
import re
import sys
from decimal import Decimal

from reader import click_odb, click
from typing import Tuple, List
from exception_codes import METAL_LAYER_ERROR, FORMAT_ERROR, NOT_FOUND_ERROR


@click.group()
def cli():
    pass


@click.command("mark_component_fixed")
@click.option(
    "-c", "--cell-name", required=True, help="Cell name of the components to mark fixed"
)
@click_odb
def mark_component_fixed(cell_name, reader):
    instances = reader.block.getInsts()
    for instance in instances:
        if instance.getMaster().getName() == cell_name:
            instance.setPlacementStatus("FIRM")


cli.add_command(mark_component_fixed)


def get_die_area(def_file, input_lefs):
    die_area_dbu = (-1, -1, -1, -1)
    db = odb.dbDatabase.create()
    for lef in input_lefs:
        odb.read_lef(db, lef)
    odb.read_def(db.getTech(), def_file)
    die_area = db.getChip().getBlock().getDieArea()
    if die_area:
        dbu = db.getChip().getBlock().getDefUnits()
        die_area_dbu = (
            die_area.xMin() / dbu,
            die_area.yMin() / dbu,
            die_area.xMax() / dbu,
            die_area.yMax() / dbu,
        )

    return die_area_dbu


def move_diearea(target_db, input_lefs, template_def):
    source_db = odb.dbDatabase.create()

    for lef in input_lefs:
        odb.read_lef(source_db, lef)
    odb.read_def(source_db.getTech(), template_def)

    assert (
        source_db.getTech().getManufacturingGrid()
        == target_db.getTech().getManufacturingGrid()
    )
    assert (
        source_db.getTech().getDbUnitsPerMicron()
        == target_db.getTech().getDbUnitsPerMicron()
    )

    diearea = source_db.getChip().getBlock().getDieArea()
    output_block = target_db.getChip().getBlock()
    output_block.setDieArea(diearea)


@click.command("move_diearea")
@click.option("-i", "--template-def", required=True, help="Input DEF")
@click_odb
def move_diearea_command(reader, input_lefs, template_def):
    """
    Move die area from input def to output def
    """
    move_diearea(reader.db, input_lefs, template_def)


def check_pin_grid(manufacturing_grid, dbu_per_microns, pin_name, pin_coordinate):
    if (pin_coordinate % manufacturing_grid) != 0:
        print(
            f"[ERROR] Pin {pin_name}'s coordinate {pin_coordinate} does not lie on the manufacturing grid.",
            file=sys.stderr,
        )  # IDK how to do this
        return True


def relocate_pins(db, input_lefs, template_def, permissive, copy_def_power=False):
    # --------------------------------
    # 1. Find list of all bterms in existing database
    # --------------------------------
    source_db = db
    source_bterms = source_db.getChip().getBlock().getBTerms()

    manufacturing_grid = source_db.getTech().getManufacturingGrid()
    dbu_per_microns = source_db.getTech().getDbUnitsPerMicron()

    print(
        f"Using manufacturing grid: {manufacturing_grid}",
        f"Using dbu per mircons: {dbu_per_microns}",
    )

    all_bterm_names = set()

    for source_bterm in source_bterms:
        source_name = source_bterm.getName()
        # TODO: Check for pin name matches net name
        # print("Bterm", source_name, "is declared as", source_bterm.getSigType())

        # --------------------------------
        # 3. Check no bterms should be marked as power, because it is assumed that caller already removed them
        # --------------------------------
        sigtype = source_bterm.getSigType()
        if sigtype in ["POWER", "GROUND"]:
            print(
                f"[WARNING] Bterm {source_name} is declared as a '{sigtype}' pin. It will be ignored.",
                file=sys.stderr,
            )
            continue
        all_bterm_names.add(source_name)

    print(
        f"Found {len(all_bterm_names)} block terminals in existing database...",
    )

    # --------------------------------
    # 2. Read the donor def
    # --------------------------------
    template_db = odb.dbDatabase.create()
    for lef in input_lefs:
        odb.read_lef(template_db, lef)
    odb.read_def(template_db.getTech(), template_def)
    template_bterms = template_db.getChip().getBlock().getBTerms()

    assert (
        source_db.getTech().getManufacturingGrid()
        == template_db.getTech().getManufacturingGrid()
    )
    assert (
        source_db.getTech().getDbUnitsPerMicron()
        == template_db.getTech().getDbUnitsPerMicron()
    )

    # --------------------------------
    # 3. Create a dict with net -> pin locations.
    # --------------------------------
    template_bterm_locations = dict()

    for template_bterm in template_bterms:
        template_name = template_bterm.getName()
        template_pins = template_bterm.getBPins()

        # TODO: Check for pin name matches net name
        for template_pin in template_pins:
            boxes = template_pin.getBoxes()

            for box in boxes:
                layer = box.getTechLayer().getName()
                if template_name not in template_bterm_locations:
                    template_bterm_locations[template_name] = []
                template_bterm_locations[template_name].append(
                    (
                        layer,
                        box.xMin(),
                        box.yMin(),
                        box.xMax(),
                        box.yMax(),
                        template_pin.getPlacementStatus(),
                    )
                )

    template_bterm_names = set(
        [
            bterm.getName()
            for bterm in template_bterms
            if bterm.getSigType() not in ["POWER", "GROUND"]
        ]
    )

    print(f"Found {len(template_bterm_locations)} template_bterms…")

    # for name in template_bterm_locations.keys():
    #     print(f"  * {name}: {template_bterm_locations[name]}")

    # --------------------------------
    # 4. Modify the pins in out def, according to dict
    # --------------------------------
    output_db = db
    output_tech = output_db.getTech()
    output_block = output_db.getChip().getBlock()
    output_bterms = output_block.getBTerms()

    if copy_def_power:
        output_bterm_names = set([bterm.getName() for bterm in output_bterms])
    else:
        output_bterm_names = set(
            [
                bterm.getName()
                for bterm in output_bterms
                if bterm.getNet().getSigType() not in ["POWER", "GROUND"]
            ]
        )
    not_in_design = template_bterm_names - output_bterm_names
    not_in_template = output_bterm_names - template_bterm_names

    mismatches_found = False
    for is_in, not_in, pins in [
        ("template", "design", not_in_design),
        ("design", "template", not_in_template),
    ]:
        for name in pins:
            mismatches_found = True
            if permissive:
                print(
                    f"[WARNING] {name} not found in {not_in} layout, but found in {is_in} layout.",
                )
            else:
                print(
                    f"[ERROR] {name} not found in {not_in} layout, but found in {is_in} layout.",
                    file=sys.stderr,
                )

    if mismatches_found and not permissive:
        exit(os.EX_DATAERR)

    if copy_def_power:
        # If asked, we copy power pins from template
        for bterm in template_bterms:
            if bterm.getSigType() not in ["POWER", "GROUND"]:
                continue
            pin_name = bterm.getName()
            pin_net_name = bterm.getNet().getName()
            pin_net = output_block.findNet(pin_net_name)
            if pin_net is None:
                pin_net = odb.dbNet.create(output_block, pin_net_name, True)
                pin_net.setSpecial()
                pin_net.setSigType(bterm.getSigType())
            pin_bterm = odb.dbBTerm.create(pin_net, pin_name)
            pin_bterm.setSigType(bterm.getSigType())
            output_bterms.append(pin_bterm)

    grid_errors = False
    for output_bterm in output_bterms:
        name = output_bterm.getName()
        output_bpins = output_bterm.getBPins()

        if name not in template_bterm_locations:
            continue

        if (name not in all_bterm_names) and not copy_def_power:
            continue

        for output_bpin in output_bpins:
            odb.dbBPin.destroy(output_bpin)

        for template_bterm_location_tuple in template_bterm_locations[name]:
            layer = output_tech.findLayer(template_bterm_location_tuple[0])

            # --------------------------------
            # 6.2 Create new pin
            # --------------------------------

            output_new_bpin = odb.dbBPin.create(output_bterm)

            print(
                f"Wrote pin {name} at layer {layer.getName()} at {template_bterm_location_tuple[1:]}..."
            )
            grid_errors = (
                check_pin_grid(
                    manufacturing_grid,
                    dbu_per_microns,
                    name,
                    template_bterm_location_tuple[1],
                )
                or grid_errors
            )
            grid_errors = (
                check_pin_grid(
                    manufacturing_grid,
                    dbu_per_microns,
                    name,
                    template_bterm_location_tuple[2],
                )
                or grid_errors
            )
            grid_errors = (
                check_pin_grid(
                    manufacturing_grid,
                    dbu_per_microns,
                    name,
                    template_bterm_location_tuple[3],
                )
                or grid_errors
            )
            grid_errors = (
                check_pin_grid(
                    manufacturing_grid,
                    dbu_per_microns,
                    name,
                    template_bterm_location_tuple[4],
                )
                or grid_errors
            )
            odb.dbBox.create(
                output_new_bpin,
                layer,
                template_bterm_location_tuple[1],
                template_bterm_location_tuple[2],
                template_bterm_location_tuple[3],
                template_bterm_location_tuple[4],
            )
            output_new_bpin.setPlacementStatus(template_bterm_location_tuple[5])

    if grid_errors:
        print(
            "[ERROR] Some pins were grid-misaligned. Please check the log.",
            file=sys.stderr,
        )
        exit(os.EX_DATAERR)


@click.command("relocate_pins")
@click.option(
    "-t",
    "--template-def",
    required=True,
    help="Template DEF to use the locations of pins from.",
)
@click_odb
def relocate_pins_command(reader, input_lefs, template_def):
    """
    Moves pins that are common between a template_def and the database to the
    location specified in the template_def.

    Assumptions:
        * The template def lacks power pins.
        * All pins are on metal layers (none on vias.)
        * All pins are rectangular.
        * All pins have unique names.
        * All pin names match the net names in the template DEF.
    """
    relocate_pins(reader.db, input_lefs, template_def)


cli.add_command(relocate_pins_command)


@click.command("remove_components")
@click.option(
    "-m",
    "--match",
    "rx_str",
    default="^.+$",
    help="Regular expression to match for components to be removed. (Default: '^.+$', matches all strings.)",
)
@click_odb
def remove_components(rx_str, reader):
    matcher = re.compile(rx_str)
    instances = reader.block.getInsts()
    for instance in instances:
        name = instance.getName()
        name_m = matcher.search(name)
        if name_m is not None:
            odb.dbInst.destroy(instance)


cli.add_command(remove_components)


@click.command("remove_nets")
@click.option(
    "-m",
    "--match",
    "rx_str",
    default="^.+$",
    help="Regular expression to match for nets to be removed. (Default: '^.+$', matches all strings.)",
)
@click.option(
    "--empty-only",
    is_flag=True,
    default=False,
    help="Adds a further condition to only remove empty nets (i.e. unconnected nets).",
)
@click_odb
def remove_nets(rx_str, empty_only, reader):
    matcher = re.compile(rx_str)
    nets = reader.block.getNets()
    for net in nets:
        name = net.getName()
        name_m = matcher.match(name)
        if name_m is not None:
            if empty_only and len(net.getITerms()) > 0:
                continue
            # BTerms = PINS, if it has a pin we need to keep the net
            if len(net.getBTerms()) > 0:
                for port in net.getITerms():
                    odb.dbITerm.disconnect(port)
            else:
                odb.dbNet.destroy(net)


cli.add_command(remove_nets)


@click.command("remove_pins")
@click.option(
    "-m",
    "--match",
    "rx_str",
    default="^.+$",
    help="Regular expression to match for components to be removed. (Default: '^.+$', matches all strings.)",
)
@click_odb
def remove_pins(rx_str, reader):
    matcher = re.compile(rx_str)
    pins = reader.block.getBTerms()
    for pin in pins:
        name = pin.getName()
        name_m = matcher.search(name)
        if name_m is not None:
            odb.dbBTerm.destroy(pin)


cli.add_command(remove_pins)


@click.command("replace_instance_prefixes")
@click.option("-f", "--original-prefix", required=True, help="The original prefix.")
@click.option("-t", "--new-prefix", required=True, help="The new prefix.")
@click_odb
def replace_instance_prefixes(original_prefix, new_prefix, reader):
    for instance in reader.block.getInsts():
        name: str = instance.getName()
        if name.startswith(f"{original_prefix}_"):
            new_name = name.replace(f"{original_prefix}_", f"{new_prefix}_")
            instance.rename(new_name)


cli.add_command(replace_instance_prefixes)


def parse_obstructions(obstructions) -> List[Tuple[str, List[int]]]:
    RE_NUMBER = r"[\-]?[0-9]+(\.[0-9]+)?"
    RE_OBS = (
        r"(?P<layer>\S+)\s+"
        + r"(?P<bbox>"
        + RE_NUMBER
        + r"\s+"
        + RE_NUMBER
        + r"\s+"
        + RE_NUMBER
        + r"\s+"
        + RE_NUMBER
        + r") *$"
    )

    obs_list = []
    for obs in obstructions:
        obs = obs.strip()
        m = re.match(RE_OBS, obs)
        if m is None:
            print(
                f"[ERROR] Incorrectly formatted input {obs}.\n Format: layer llx lly urx ury, ...",
                file=sys.stderr,
            )
            sys.exit(FORMAT_ERROR)
        else:
            layer = m.group("layer")
            bbox = [Decimal(x) for x in m.group("bbox").split()]
            obs_list.append((layer, bbox))

    return obs_list


@click.command("add_obstructions")
@click.option(
    "-O",
    "--obstructions",
    multiple=True,
    required=True,
    help="Format: layer llx lly urx ury, (microns)",
)
@click_odb
def add_obstructions(reader, input_lefs, obstructions):
    obs_list = parse_obstructions(obstructions)
    for obs in obs_list:
        layer = obs[0]
        odb_layer = reader.tech.findLayer(layer)
        if odb_layer is None:
            print(f"[ERROR] layer {layer} doesn't exist.", file=sys.stderr)
            sys.exit(METAL_LAYER_ERROR)
        bbox = obs[1]
        dbu = reader.tech.getDbUnitsPerMicron()
        bbox = [int(x * dbu) for x in bbox]
        print(f"Creating an obstruction on {layer} at {bbox} (DBU)…")
        odb.dbObstruction_create(reader.block, reader.tech.findLayer(layer), *bbox)


cli.add_command(add_obstructions)


@click.command("remove_obstructions")
@click.option(
    "-O",
    "--obstructions",
    multiple=True,
    required=True,
    help="Format: layer llx lly urx ury, (microns)",
)
@click_odb
def remove_obstructions(reader, input_lefs, obstructions):
    dbu: int = reader.tech.getDbUnitsPerMicron()
    existing_obstructions: List[Tuple[str, List[int], odb.dbObstruction]] = []

    for odb_obstruction in reader.block.getObstructions():
        bbox = odb_obstruction.getBBox()
        existing_obstructions.append(
            (
                bbox.getTechLayer().getName(),
                [
                    bbox.xMin(),
                    bbox.yMin(),
                    bbox.xMax(),
                    bbox.yMax(),
                ],
                odb_obstruction,
            )
        )

    for obs in parse_obstructions(obstructions):
        layer, bbox = obs
        bbox = [int(x * dbu) for x in bbox]  # To dbus
        found = False
        if reader.tech.findLayer(layer) is None:
            print(f"[ERROR] layer {layer} doesn't exist.", file=sys.stderr)
            sys.exit(METAL_LAYER_ERROR)
        for odb_obstruction in existing_obstructions:
            odb_layer, odb_bbox, odb_obj = odb_obstruction
            if (odb_layer, odb_bbox) == (layer, bbox):
                print(f"Removing obstruction on {layer} at {bbox} (DBU)…")
                found = True
                odb.dbObstruction_destroy(odb_obj)
            if found:
                break
        if not found:
            print(
                f"[ERROR] Obstruction on {layer} at {bbox} (DBU) not found.",
                file=sys.stderr,
            )
            sys.exit(NOT_FOUND_ERROR)


cli.add_command(remove_obstructions)

if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# Copyright 2020 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from decimal import Decimal

from reader import click_odb, click


@click.group
def cli():
    pass


LEF2OA_MAP = {
    "N": "R0",
    "S": "R180",
    "W": "R90",
    "E": "R270",
    "FN": "MY",
    "FS": "MX",
    "FW": "MXR90",
    "FE": "MYR90",
}


def lef_rot_to_oa_rot(rot):
    if rot in LEF2OA_MAP:
        return LEF2OA_MAP[rot]
    else:
        assert rot in [item[1] for item in LEF2OA_MAP.items()], rot
        return rot


def gridify(n, f):
    """
    e.g., (1.1243, 0.005) -> 1.120
    """
    return round(n / f) * f


@click.command()
@click.option("-c", "--config", required=True, help="Configuration file")
@click.option(
    "-f",
    "--fixed",
    default=False,
    is_flag=True,
    help="A flag to signal whether the placement should be fixed or not",
)
@click_odb
def manual_macro_placement(reader, config, fixed):
    """
    Places macros in positions and orientations specified by a config file
    """

    db_units_per_micron = reader.block.getDbUnitsPerMicron()

    # read config
    macros = {}
    with open(config, "r") as config_file:
        for line in config_file:
            # Discard comments and empty lines
            line = line.split("#")[0].strip()
            if not line:
                continue
            line = line.split()
            name, x, y, orientation = line
            macro_data = [
                name,
                int(Decimal(x) * db_units_per_micron),
                int(Decimal(y) * db_units_per_micron),
                orientation,
            ]
            name_escaped = reader.escape_verilog_name(name)
            macros[name_escaped] = macro_data

    print("Placing the following macros:")
    print(macros)

    print("Design name:", reader.name)

    macros_cnt = len(macros)
    for inst in reader.block.getInsts():
        inst_name = inst.getName()
        if inst.isFixed():
            assert inst_name not in macros, inst_name
            continue
        if inst_name in macros:
            print("Placing", inst_name)
            macro_data = macros[inst_name]
            _, x, y, orientation = macro_data
            x = gridify(x, 5)
            y = gridify(y, 5)
            inst.setOrient(lef_rot_to_oa_rot(orientation))
            inst.setLocation(x, y)
            if fixed:
                inst.setPlacementStatus("FIRM")
            else:
                inst.setPlacementStatus("PLACED")
            del macros[inst_name]

    if len(macros):
        print("Declared macros not instantiated in design:", file=sys.stderr)
        for macro in macros.values():
            print(f"* {macro[0]}", file=sys.stderr)
        exit(1)

    print(f"Successfully placed {macros_cnt} instances.")


cli.add_command(manual_macro_placement)


@click.command()
@click_odb
def manual_global_placement(reader):
    db_units_per_micron = reader.block.getDbUnitsPerMicron()

    data = reader.config["MANUAL_GLOBAL_PLACEMENTS"]
    not_found = []
    for instance, info in data.items():
        name_escaped = reader.escape_verilog_name(instance)
        x, y = info["location"]
        orientation = lef_rot_to_oa_rot(info["orientation"])
        found = False
        for inst in reader.block.getInsts():
            if inst.getName() == name_escaped:
                found = True
                x_dbu = int(x * db_units_per_micron)
                y_dbu = int(y * db_units_per_micron)
                inst.setOrient(lef_rot_to_oa_rot(orientation))
                inst.setLocation(x_dbu, y_dbu)
                inst.setPlacementStatus("PLACED")
                break
        if not found:
            not_found.append(instance)

    if len(not_found):
        print(
            "[ERROR] One or more instances not found. Make sure you use their Verilog and not their LEF names."
        )
        for instance in not_found:
            print(f"* {instance}")
        exit(1)


cli.add_command(manual_global_placement)

if __name__ == "__main__":
    cli()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from math import inf
from decimal import Decimal
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Mapping, Optional, Tuple, ClassVar, Dict


from ..types import Number, is_number, is_real_number

MetricAggregator = Tuple[Number, Callable[[Iterable[Number]], Number]]

sum_aggregator: MetricAggregator = (0, lambda x: sum(x))
min_aggregator: MetricAggregator = (inf, min)
max_aggregator: MetricAggregator = (-inf, max)


@dataclass
class MetricComparisonResult:
    """
    :param metric_name: The name of the metric that has been compared
    :param gold: The "gold" value being compared against
    :param new: The new value being evaluated
    :param delta: ``None`` if and only if ``before`` - ``after`` is an invalid number.
        Evaluates to ``after - before``\\.
    :param delta_pct: ``None`` if ``delta`` is None or before is zero.
        Otherwise, evaluates to ``delta / before * 100``\\.
    :param better: Whether the change in the value is considered a good thing or
        not. ``None`` if ``delta`` is None or has no value set for
        ``Metric.higher_is_better``\\.
    :param critical: Whether this change of value very likely results in a dead
        chip, i.e., an increase in DRC values, or an inexplicable change in
        the number of I/O pins.

    """

    metric_name: str
    gold: Any
    new: Any
    delta: Optional[Number]
    delta_pct: Optional[Number]
    better: Optional[bool]
    critical: bool
    significant_figures: Optional[int]

    def is_changed(self) -> bool:
        return (self.delta is not None and self.delta != 0) or self.gold != self.new

    def format_values(self) -> Tuple[str, str, str]:
        before_str = str(self.gold)
        if isinstance(self.gold, float) or isinstance(self.gold, Decimal):
            before_str = str(f"{self.gold:.{self.significant_figures}f}")

        after_str = str(self.new)
        if isinstance(self.new, float) or isinstance(self.new, Decimal):
            after_str = str(f"{self.new:.{self.significant_figures}f}")

        delta_str = "N/A"
        if self.delta is not None:
            delta_str = str(round(self.delta, 6))
            if isinstance(self.delta, float) or isinstance(self.delta, Decimal):
                delta_str = str(f"{self.delta:.{self.significant_figures}f}")
            if self.delta_pct is not None:
                delta_pct_str = str(f"{self.delta_pct:.{self.significant_figures}f}")
                if self.delta_pct > 0:
                    delta_pct_str = f"+{delta_pct_str}"
                delta_str = f"{delta_str} ({delta_pct_str}%)"

        return before_str, after_str, delta_str


@dataclass
class Metric(object):
    """
    An object storing data about a metric as defined in METRICS2.1.

    :param name: The string name of the metric.
    :param aggregator: A tuple of:
        - A starting value for an accumulator
        - A reduction function

        The aim is the ability to aggregate values from various sub-metrics,
        i.e., for the metric ``timing__hold_vio__count``, the sub-metrics:

        - ``timing__hold_vio__count__corner:A``
        - ``timing__hold_vio__count__corner:B``

        Would be summed up to generate the value for ``timing__hold_vio__count``.
    :param higher_is_better: At a high level, whether a higher numeric value for
        this metric is considered "good" (such as: better utilization) or "bad"
        (such as: more antenna violations.)
    :param critical: A critical metric is always watched for any change.

    """

    name: str
    aggregator: Optional[MetricAggregator] = None
    higher_is_better: Optional[bool] = None
    dont_aggregate: Optional[Iterable[str]] = None
    critical: bool = False

    by_name: ClassVar[Dict[str, "Metric"]] = {}

    def __post_init__(self):
        Metric.by_name[self.name] = self

    def modified_name(self, modifiers: Mapping[str, str]) -> str:
        """
        :param modifiers: Modifiers of a metric (i.e. the elements postfixed to the metric in the format {key}:{value})
        :returns: The name with the modifiers added
        """
        return "__".join([self.name] + [f"{k}:{v}" for k, v in modifiers.items()])

    def compare(
        self,
        gold: Any,
        new: Any,
        significant_figures: int,
        modifiers: Optional[Mapping[str, str]] = None,
    ) -> MetricComparisonResult:
        """
        :param gold: The "gold-standard" value for this metric to compare against
        :param new: The new value for this metric being evaluated
        :param modifier: The modifiers that were parsed from the metric name
            (if applicable)- used to set the ``metric_name`` property of
            :class:`MetricComparisonResult`.
        :returns: The result of comparing two values for this metric.
        """
        is_better = None
        is_critical = self.critical and (gold != new)
        delta = None
        delta_pct = None

        if modifiers is None:
            modifiers = {}

        if is_real_number(gold) and is_real_number(new):
            if isinstance(gold, float) or isinstance(new, float):
                gold = Decimal(gold)
                new = Decimal(new)
            delta = new - gold

            if gold == 0:
                if new == 0:
                    delta_pct = Decimal(0)
            else:
                delta_pct = Decimal((delta / gold) * 100)
                if delta_pct == 0:
                    delta_pct = Decimal(0)  # Fix negative zero

            if self.higher_is_better is not None:
                if self.higher_is_better:
                    is_better = delta >= 0
                else:
                    is_better = delta <= 0
        elif self.higher_is_better is not None:
            if not is_number(gold):
                raise TypeError(
                    f"'{gold}' is not a number for metric {self.name} with non-None 'higher_is_better' field"
                )
            if not is_number(new):
                raise TypeError(
                    f"'{new}' is not a number  for metric {self.name} with non-None 'higher_is_better' field"
                )

        return MetricComparisonResult(
            self.modified_name(modifiers),
            gold,
            new,
            delta,
            delta_pct,
            is_better,
            is_critical,
            significant_figures,
        )



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The Configuration Module
------------------------

This modules includes various functions for importing and/or generating OpenLane
configuration objects. Configuration objects are the primary input to a flow.
"""
from .preprocessor import Keys
from .variable import Instance, Macro, Variable
from .config import (
    Meta,
    Config,
    InvalidConfig,
    AnyConfig,
    AnyConfigs,
    PassedDirectoryError,
    UnknownExtensionError,
)
from .flow import flow_common_variables as universal_flow_config_variables



# Copyright 2020-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
import odb

import os
import re
import sys
import math
import click
import random
from decimal import Decimal

from reader import click_odb
import ioplace_parser


def grid_to_tracks(origin, count, step):
    tracks = []
    pos = origin
    for _ in range(count):
        tracks.append(pos)
        pos += step
    assert len(tracks) > 0
    tracks.sort()

    return tracks


def equally_spaced_sequence(side, side_pin_placement, possible_locations):
    virtual_pin_count = 0
    actual_pin_count = len(side_pin_placement)
    total_pin_count = actual_pin_count + virtual_pin_count
    for i in range(len(side_pin_placement)):
        if isinstance(
            side_pin_placement[i], int
        ):  # This is an int value indicating virtual pins
            virtual_pin_count = virtual_pin_count + side_pin_placement[i]
            actual_pin_count = (
                actual_pin_count - 1
            )  # Decrement actual pin count, this value was only there to indicate virtual pin count
            total_pin_count = actual_pin_count + virtual_pin_count
    result = []
    tracks = len(possible_locations)

    if total_pin_count > tracks:
        print(
            f"[ERROR] The {side} side of the floorplan doesn't have enough slots for all the pins: {total_pin_count} pins/{tracks} slots.",
            file=sys.stderr,
        )
        print(
            "[INFO] Try re-assigning pins to other sides or making the floorplan larger.",
            file=sys.stderr,
        )
        sys.exit(1)
    elif total_pin_count == tracks:
        return possible_locations, side_pin_placement  # All positions.
    elif total_pin_count == 0:
        return result, side_pin_placement

    # From this point, pin_count always < tracks.
    tracks_per_pin = math.floor(tracks / total_pin_count)  # >=1
    # O| | | O| | | O| | |
    # tracks_per_pin = 3
    # notice the last two tracks are unused
    # thus:
    used_tracks = tracks_per_pin * (total_pin_count - 1) + 1
    unused_tracks = tracks - used_tracks

    # Place the pins at those tracks...
    current_track = unused_tracks // 2  # So that the tracks used are centered
    starting_track_index = current_track
    if virtual_pin_count == 0:  # No virtual pins
        for _ in range(0, total_pin_count):
            result.append(possible_locations[current_track])
            current_track += tracks_per_pin
    else:  # There are virtual pins
        for i in range(len(side_pin_placement)):
            if not isinstance(side_pin_placement[i], int):  # We have an actual pin
                result.append(possible_locations[current_track])
                current_track += tracks_per_pin
            else:  # Virtual Pins, so just leave their needed spaces
                current_track += tracks_per_pin * side_pin_placement[i]
        side_pin_placement = [
            pin for pin in side_pin_placement if not isinstance(pin, int)
        ]  # Remove the virtual pins from the side_pin_placement list

    print(f"Placement details for the {side} side")
    print("Virtual pin count: ", virtual_pin_count)
    print("Actual pin count: ", actual_pin_count)
    print("Total pin count: ", total_pin_count)
    print("Tracks count: ", len(possible_locations))
    print("Tracks per pin: ", tracks_per_pin)
    print("Used tracks count: ", used_tracks)
    print("Unused track count: ", unused_tracks)
    print("Starting track index: ", starting_track_index)

    VISUALIZE_PLACEMENT = False
    if VISUALIZE_PLACEMENT:
        print("Placement Map:")
        print("[", end="")
        used_track_indices = []
        for i, location in enumerate(possible_locations):
            if location in result:
                print(f"\033[91m{location}\033[0m, ", end="")
                used_track_indices.append(i)
            else:
                print(f"{location}, ", end="")
        print("]")
        print(f"Indices of used tracks: {used_track_indices}")
        print("---")

    return result, side_pin_placement


identifiers = re.compile(r"\b[A-Za-z_][A-Za-z_0-9]*\b")
standalone_numbers = re.compile(r"\b\d+\b")
trash = re.compile(r"^[^\w\d]+$")


def sorter(bterm, order: ioplace_parser.Order):
    text: str = bterm.getName()
    keys = []
    priority_keys = []
    # tokenize and add to key
    while trash.match(text) is None:
        if match := identifiers.search(text):
            bus = match[0]
            start, end = match.span(0)
            if order == ioplace_parser.Order.busMajor:
                priority_keys.append(bus)
            else:
                keys.append(bus)
            text = text[:start] + text[end + 1 :]
        elif match := standalone_numbers.search(text):
            index = int(match[0])
            if order == ioplace_parser.Order.bitMajor:
                priority_keys.append(index)
            else:
                keys.append(index)
            text = text[: match.pos] + text[match.endpos + 1 :]
        else:
            break
    return [priority_keys, keys]


@click.command()
@click.option(
    "-u",
    "--unmatched-error",
    type=click.Choice(["none", "unmatched_design", "unmatched_cfg", "both"]),
    default=True,
    help="Treat unmatched pins as error",
)
@click.option(
    "-c",
    "--config",
    required=True,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
    help="Input configuration file",
)
@click.option(
    "-v",
    "--ver-length",
    default=None,
    type=float,
    help="Length for pins with N/S orientations in microns.",
)
@click.option(
    "-h",
    "--hor-length",
    default=None,
    type=float,
    help="Length for pins with E/S orientations in microns.",
)
@click.option(
    "-V",
    "--ver-layer",
    required=True,
    help="Name of metal layer to place vertical pins on.",
)
@click.option(
    "-H",
    "--hor-layer",
    required=True,
    help="Name of metal layer to place horizontal pins on.",
)
@click.option(
    "--hor-extension",
    default=0,
    type=float,
    help="Extension for vertical pins in microns.",
)
@click.option(
    "--ver-extension",
    default=0,
    type=float,
    help="Extension for horizontal pins in microns.",
)
@click.option(
    "--ver-width-mult", default=2, type=float, help="Multiplier for vertical pins."
)
@click.option(
    "--hor-width-mult", default=2, type=float, help="Multiplier for horizontal pins."
)
@click_odb
def io_place(
    reader,
    config,
    ver_layer,
    hor_layer,
    ver_width_mult,
    hor_width_mult,
    hor_length,
    ver_length,
    hor_extension,
    ver_extension,
    unmatched_error,
):
    """
    Places the IOs in an input def with an optional config file that supports regexes.

    Config format:
    #N|#S|#E|#W
    pin1_regex (low co-ordinates to high co-ordinates; e.g., bottom to top and left to right)
    pin2_regex
    ...

    #S|#N|#E|#W
    """
    config_file_name = config
    micron_in_units = reader.dbunits

    H_EXTENSION = int(micron_in_units * hor_extension)
    V_EXTENSION = int(micron_in_units * ver_extension)

    if H_EXTENSION < 0:
        H_EXTENSION = 0

    if V_EXTENSION < 0:
        V_EXTENSION = 0

    H_LAYER = reader.tech.findLayer(hor_layer)
    V_LAYER = reader.tech.findLayer(ver_layer)

    H_WIDTH = int(Decimal(hor_width_mult) * H_LAYER.getWidth())
    V_WIDTH = int(Decimal(ver_width_mult) * V_LAYER.getWidth())

    if hor_length is not None:
        H_LENGTH = int(micron_in_units * hor_length)
    else:
        H_LENGTH = max(
            int(
                math.ceil(
                    H_LAYER.getArea() * micron_in_units * micron_in_units / H_WIDTH
                )
            ),
            H_WIDTH,
        )

    if ver_length is not None:
        V_LENGTH = int(micron_in_units * ver_length)
    else:
        V_LENGTH = max(
            int(
                math.ceil(
                    V_LAYER.getArea() * micron_in_units * micron_in_units / V_WIDTH
                )
            ),
            V_WIDTH,
        )

    # read config + calculate minima
    config_file_str = open(config_file_name, "r", encoding="utf8").read()

    try:
        info_by_side = ioplace_parser.parse(config_file_str)
    except ValueError as e:
        print(f"An exception occurred: {e}")
        exit(os.EX_DATAERR)

    print("Top-level design name:", reader.name)

    bterms = [
        bterm
        for bterm in reader.block.getBTerms()
        if bterm.getSigType() not in ["POWER", "GROUND"]
    ]

    for side, side_info in info_by_side.items():
        min = (
            (V_WIDTH + V_LAYER.getSpacing())
            if side in ["N", "S"]
            else (H_WIDTH + H_LAYER.getSpacing())
        ) / reader.dbunits
        if side_info.min_distance is None:
            side_info.min_distance = min
        if side_info.min_distance < min:
            print(
                f"[WARNING] Overriding minimum distance {side_info.min_distance} with {min} for pins on side {side} to avoid overlap.",
                file=sys.stderr,
            )
            side_info.min_distance = min

    # build a list of pins
    pin_placement = {"N": [], "E": [], "W": [], "S": []}

    regex_by_bterm = {}
    unmatched_regexes = set()
    for side, side_info in info_by_side.items():
        for pin in side_info.pins:
            if isinstance(pin, int):  # Virtual pins
                pin_placement[side].append(pin)
                continue

            anchored_regex = f"^{pin}$"  # anchor
            matched = False
            collected = []
            for bterm in bterms:
                pin_name = bterm.getName()
                if re.match(anchored_regex, pin_name) is None:
                    continue
                if bterm in regex_by_bterm:
                    print(
                        f"[ERROR] Multiple regexes matched {pin_name}. Those are {regex_by_bterm[bterm]} and {pin}",
                        file=sys.stderr,
                    )
                    sys.exit(os.EX_DATAERR)
                regex_by_bterm[bterm] = pin
                collected.append(bterm)
                matched = True
            collected.sort(key=partial(sorter, order=side_info.sort_mode))
            pin_placement[side] += collected
            if not matched:
                unmatched_regexes.add(pin)

    # check for extra or missing pins
    not_in_design = unmatched_regexes
    not_in_config = set(
        [bterm.getName() for bterm in bterms if bterm not in regex_by_bterm]
    )
    mismatches_found = False
    for is_in, not_in, pins in [
        ("config", "design", not_in_design),
        ("design", "config", not_in_config),
    ]:
        for name in pins:
            if (
                is_in == "config"
                and (unmatched_error in {"unmatched_cfg", "both"})
                or is_in == "design"
                and (unmatched_error in {"unmatched_design", "both"})
            ):
                mismatches_found = True
                print(
                    f"[ERROR] {name} not found in {not_in} but found in {is_in}.",
                    file=sys.stderr,
                )
            else:
                print(
                    f"[WARNING] {name} not found in {not_in} but found in {is_in}.",
                    file=sys.stderr,
                )

    if mismatches_found:
        print("Critical mismatches found.")
        exit(os.EX_DATAERR)

    if len(not_in_config) > 0:
        print("Assigning random sides to unmatched pins…")
        for bterm in not_in_config:
            random_side = random.choice(list(pin_placement.keys()))
            pin_placement[random_side].append(bterm)

    # generate slots
    DIE_AREA = reader.block.getDieArea()
    BLOCK_LL_X = DIE_AREA.xMin()
    BLOCK_LL_Y = DIE_AREA.yMin()
    BLOCK_UR_X = DIE_AREA.xMax()
    BLOCK_UR_Y = DIE_AREA.yMax()

    print("Block boundaries:", BLOCK_LL_X, BLOCK_LL_Y, BLOCK_UR_X, BLOCK_UR_Y)

    origin, count, h_step = reader.block.findTrackGrid(H_LAYER).getGridPatternY(0)
    print(f"Horizontal Tracks Origin: {origin}, Count: {count}, Step: {h_step}")
    h_tracks = grid_to_tracks(origin, count, h_step)

    origin, count, v_step = reader.block.findTrackGrid(V_LAYER).getGridPatternX(0)
    print(f"Vertical Tracks Origin: {origin}, Count: {count}, Step: {v_step}")
    v_tracks = grid_to_tracks(origin, count, v_step)

    pin_tracks = {}
    for side in pin_placement:
        if side in ["N", "S"]:
            min_distance = info_by_side[side].min_distance * micron_in_units
            pin_tracks[side] = [
                v_tracks[i]
                for i in range(len(v_tracks))
                if (i % (math.ceil(min_distance / v_step))) == 0
            ]
        elif side in ["W", "E"]:
            pin_tracks[side] = [
                h_tracks[i]
                for i in range(len(h_tracks))
                if (
                    i
                    % (
                        math.ceil(
                            info_by_side[side].min_distance * micron_in_units / h_step
                        )
                    )
                )
                == 0
            ]

    # reversals (including randomly-assigned pins, if needed)
    for side, side_info in info_by_side.items():
        if side_info.reverse_result:
            pin_placement[side].reverse()

    # create the pins
    for side in pin_placement:
        slots, pin_placement[side] = equally_spaced_sequence(
            side, pin_placement[side], pin_tracks[side]
        )

        assert len(slots) == len(pin_placement[side])

        for i in range(len(pin_placement[side])):
            bterm = pin_placement[side][i]
            slot = slots[i]
            pin_name = bterm.getName()
            pins = bterm.getBPins()
            if len(pins) > 0:
                print(
                    f"[WARNING] {pin_name} already has shapes. The shapes will be modified.",
                    file=sys.stderr,
                )
                assert len(pins) == 1
                pin_bpin = pins[0]
            else:
                pin_bpin = odb.dbBPin_create(bterm)

            pin_bpin.setPlacementStatus("PLACED")

            if side in ["N", "S"]:
                rect = odb.Rect(0, 0, V_WIDTH, V_LENGTH + V_EXTENSION)
                if side == "N":
                    y = BLOCK_UR_Y - V_LENGTH
                else:
                    y = BLOCK_LL_Y - V_EXTENSION
                rect.moveTo(slot - V_WIDTH // 2, y)
                odb.dbBox_create(pin_bpin, V_LAYER, *rect.ll(), *rect.ur())
            else:
                rect = odb.Rect(0, 0, H_LENGTH + H_EXTENSION, H_WIDTH)
                if side == "E":
                    x = BLOCK_UR_X - H_LENGTH
                else:
                    x = BLOCK_LL_X - H_EXTENSION
                rect.moveTo(x, slot - H_WIDTH // 2)
                odb.dbBox_create(pin_bpin, H_LAYER, *rect.ll(), *rect.ur())


if __name__ == "__main__":
    io_place()



# Copyright 2021-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odb

import re
from typing import List

from reader import click_odb, click


def get_pin_name(pin: odb.dbITerm):
    cell = pin.getInst()
    cell_name = cell.getName()
    master_pin = pin.getMTerm()
    master_pin_name = master_pin.getConstName()
    return f"{cell_name}/{master_pin_name}"


def get_sinks_terms(net: odb.dbNet) -> List[odb.dbITerm]:
    sinks = []
    for it in net.getITerms():
        cell = it.getInst()
        cell_pin = it.getMTerm()

        master_instance = cell.getMaster()
        master_name = master_instance.getConstName()

        if cell_pin.getIoType() == "INPUT":
            print(
                f"  * Net {net.getConstName()} sinks into {get_pin_name(it)} ({master_name})..."
            )
            sinks.append(it)

    return sinks


def get_drivers(net: odb.dbNet) -> List[odb.dbInst]:
    drivers = []
    for it in net.getITerms():
        cell = it.getInst()
        cell_pin = it.getMTerm()

        master_instance = cell.getMaster()
        master_name = master_instance.getConstName()

        if cell_pin.getIoType() == "OUTPUT":
            print(
                f"  * Net {net.getConstName()} is driven by {get_pin_name(it)} ({master_name})..."
            )
            drivers.append(cell)

    return drivers


def get_io(cell: odb.dbInst):
    inputs = []
    outputs = []
    for iterm in cell.getITerms():
        if iterm.isSpecial():  # Skip special nets
            continue

        if iterm.isInputSignal():
            input_pin_net = iterm.getNet()
            inputs.append((iterm, input_pin_net))

        if iterm.isOutputSignal():
            output_pin_net = iterm.getNet()
            outputs.append((iterm, output_pin_net))

    return inputs, outputs


@click.command()
@click.option(
    "-m",
    "--match",
    "rx_str",
    required=True,
    help="A regular expression matching all nets to remove.",
)
@click_odb
def remove_buffers(reader, rx_str):
    if rx_str != "^$":
        # Save some compute time :)

        rx = re.compile(rx_str)

        design_nets = reader.block.getNets()
        dont_buffer_nets = [
            net for net in design_nets if rx.match(net.getConstName()) is not None
        ]

        for net in dont_buffer_nets:
            net_name = net.getConstName()
            print(f"* Attempting to unbuffer {net_name}...")
            # get the cells driving the dont buffer net
            drivers = get_drivers(net)

            if len(drivers) > 1:
                print(f"Net {net_name} is driven by multiple cells.")
                exit(1)
            elif len(drivers) == 0:
                print(f"Net {net_name} is not driven by any cell..")
                exit(1)

            buffer = drivers[0]
            buffer_name = drivers[0].getName()
            master = buffer.getMaster()
            master_name = master.getConstName()

            if "buf" not in master_name:
                print(
                    f"*  {net_name} isn't driven by a buffer cell. It is driven by {buffer_name} ({master_name}). Skipping..."
                )
                continue

            # get the net connected to the input pin of this buffer
            inputs, outputs = get_io(buffer)

            if len(inputs) != 1:
                print(
                    f"*  {master_name} has more than one output port. Doesn't appear to actually be a buffer. Skipping..."
                )
                continue
            if len(outputs) != 1:
                print(
                    f"{master_name} has more than one output port. Doesn't appear to actually be a buffer. Skipping..."
                )
                continue

            _, input_net = inputs[0]
            _, output_net = outputs[0]

            print("  * Reconnecting IO...")
            # We connect the driver's output to the output_net: there may not be a
            # sink with ITerms in case of things like output ports for example.
            buffer_input_driver = get_drivers(input_net)[0]
            _, bid_outputs = get_io(buffer_input_driver)
            (bid_iterm, _) = bid_outputs[0]
            bid_iterm.connect(output_net)
            print(
                f"  * Connected buffer output({output_net.getConstName()}) to {get_pin_name(bid_iterm)}."
            )

            print(f"  * Removing buffer {buffer_name} ({master_name})...")
            odb.dbInst.destroy(buffer)

            input_net_sinks = get_sinks_terms(input_net)
            for iterm in input_net_sinks:
                iterm.connect(output_net)
                print(
                    f"  * Connected buffer output({output_net.getConstName()}) to {get_pin_name(iterm)}."
                )

            print(f"  * Removing net {input_net.getName()}...")
            odb.dbNet.destroy(input_net)

        print("  * Done.")


if __name__ == "__main__":
    remove_buffers()
